from collections import OrderedDict
from typing import Dict, List, Optional, Union

from phidata.aws.args import AwsArgs
from phidata.app.base_app import BaseApp
from phidata.app.phidata_app import PhidataApp
from phidata.aws.api_client import AwsApiClient
from phidata.aws.resource.base import AwsResource
from phidata.aws.resource.group import AwsResourceGroup
from phidata.aws.resource.utils import (
    filter_and_flatten_aws_resource_groups,
)
from phidata.utils.cli_console import (
    print_info,
    print_heading,
    confirm_yes_no,
)
from phidata.utils.log import logger


class AwsWorker:
    """This class interacts with the Aws API."""

    def __init__(self, aws_args: AwsArgs) -> None:
        self.aws_args: AwsArgs = aws_args
        self.aws_api_client: AwsApiClient = AwsApiClient(
            aws_region=self.aws_args.aws_region,
            aws_profile=self.aws_args.aws_profile,
        )
        logger.debug(f"**-+-** AwsWorker created")

    def is_client_initialized(self) -> bool:
        return self.aws_api_client.is_initialized()

    def are_resources_active(self) -> bool:
        # TODO: fix this
        return False

    ######################################################
    ## Build Resources
    ######################################################

    def build_aws_resource_groups(
        self,
        app_filter: Optional[str] = None,
    ) -> Optional[Dict[str, AwsResourceGroup]]:
        """
        Build the AwsResourceGroups for the requested apps
        """
        logger.debug("-*- Initializing AwsResourceGroups")

        aws_resource_groups: Optional[Dict[str, AwsResourceGroup]] = None

        aws_apps: Optional[List[Union[BaseApp, PhidataApp]]] = self.aws_args.apps
        aws_rgs: Optional[List[AwsResourceGroup]] = self.aws_args.resources

        num_apps = len(aws_apps) if aws_apps is not None else 0
        num_rgs = len(aws_rgs) if aws_rgs is not None else 0
        num_rgs_to_build = num_apps + num_rgs
        num_rgs_built = 0

        # Step 1: Convert each PhidataApp to AwsResourceGroup.
        if aws_apps is not None and isinstance(aws_apps, list):
            for app in aws_apps:
                if app.args is None:
                    logger.error("Args for App {} are None".format(app))
                    num_rgs_built += 1
                    continue

                if not app.enabled:
                    logger.debug(f"{app.name} disabled")
                    num_rgs_built += 1
                    continue

                # skip apps not matching app_filter if provided
                if app_filter is not None:
                    if app_filter.lower() not in app.name:
                        logger.debug(f"Skipping {app.name}")
                        num_rgs_built += 1
                        continue

                logger.debug("-*- App: {}".format(app.name))

                ######################################################################
                # NOTE: VERY IMPORTANT TO GET RIGHT
                # Pass down parameters from K8sArgs -> PhidataApp
                # The K8sConfig inherits these params from the WorkspaceConfig
                # 1. Pass down the paths from the WorkspaceConfig
                # 2. Pass down k8s_env
                # 3. Pass down common cloud configuration. eg: aws_region, aws_profile
                # This should match phidata.infra.prep_infra_config.prep_infra_config()
                ######################################################################

                # -*- Path parameters
                app.scripts_dir = self.aws_args.scripts_dir
                app.storage_dir = self.aws_args.storage_dir
                app.meta_dir = self.aws_args.meta_dir
                app.products_dir = self.aws_args.products_dir
                app.notebooks_dir = self.aws_args.notebooks_dir
                app.workflows_dir = self.aws_args.workflows_dir
                # The ws_root_path is the ROOT directory for the workspace
                app.workspace_root_path = self.aws_args.workspace_root_path
                app.workspace_config_dir = self.aws_args.workspace_config_dir
                app.workspace_config_file_path = (
                    self.aws_args.workspace_config_file_path
                )

                # -*- AWS parameters
                # only update the params if they are not available on the app.
                # so we can prefer the app param if provided
                if app.aws_region is None and self.aws_args.aws_region is not None:
                    app.aws_region = self.aws_args.aws_region
                if app.aws_profile is None and self.aws_args.aws_profile is not None:
                    app.aws_profile = self.aws_args.aws_profile
                if (
                    app.aws_config_file is None
                    and self.aws_args.aws_config_file is not None
                ):
                    app.aws_config_file = self.aws_args.aws_config_file
                if (
                    app.aws_shared_credentials_file is None
                    and self.aws_args.aws_shared_credentials_file is not None
                ):
                    app.aws_shared_credentials_file = (
                        self.aws_args.aws_shared_credentials_file
                    )

                app_rgs: Optional[
                    Dict[str, AwsResourceGroup]
                ] = app.get_aws_resource_groups(aws_build_context=self.aws_args)
                if app_rgs is not None:
                    if aws_resource_groups is None:
                        aws_resource_groups = OrderedDict()
                    aws_resource_groups.update(app_rgs)
                    num_rgs_built += 1

        # Step 2: Add all AwsResourceGroups to the aws_resource_groups dict
        if aws_rgs is not None and isinstance(aws_rgs, list):
            for rg in aws_rgs:
                if not rg.enabled:
                    logger.debug(f"{rg.name} disabled")
                    num_rgs_built += 1
                    continue

                # skip groups not matching app_filter if provided
                if app_filter is not None:
                    if app_filter.lower() not in rg.name:
                        logger.debug(f"Skipping {rg.name}")
                        num_rgs_built += 1
                        continue

                if isinstance(rg, AwsResourceGroup):
                    if aws_resource_groups is None:
                        aws_resource_groups = OrderedDict()
                    aws_resource_groups[rg.name] = rg
                    num_rgs_built += 1

        logger.debug(f"# AwsResourceGroups built: {num_rgs_built}/{num_rgs_to_build}")
        return aws_resource_groups

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Creating AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return True

        aws_resources_to_create: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        # Validate resources to be created
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in aws_resources_to_create:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {len(aws_resources_to_create)} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping deploy")
                print_info("-*-")
                exit(0)

        # track the total number of AwsResources to create for validation
        num_resources_to_create: int = len(aws_resources_to_create)
        num_resources_created: int = 0

        for resource in aws_resources_to_create:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_created = resource.create(aws_client=self.aws_api_client)
                if _resource_created:
                    num_resources_created += 1
                    print_info("Resource created")
                else:
                    logger.error(
                        f"Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                    )
                    if not self.aws_args.continue_on_create_failure:
                        return False
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources created: {num_resources_created}/{num_resources_to_create}"
        )
        if num_resources_to_create == num_resources_created:
            return True

        logger.error(
            f"Resources created: {num_resources_created} do not match resources required: {num_resources_to_create}"
        )
        return False

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Creating AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return

        aws_resources_to_create: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        num_resources_to_create: int = len(aws_resources_to_create)
        print_heading("--**-- Aws resources:")
        for resource in aws_resources_to_create:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_create} resources")

    ######################################################
    ## Get Resources
    ######################################################

    def get_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[AwsResource]]:
        logger.debug("-*- Getting AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return None

        aws_resources: List[AwsResource] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        return aws_resources

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Deleting AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return True

        aws_resources_to_delete: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            sort_order="delete",
        )

        # Validate resources to be deleted
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in aws_resources_to_delete:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {len(aws_resources_to_delete)} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping delete")
                print_info("-*-")
                exit(0)

        # track the total number of AwsResources to delete for validation
        num_resources_to_delete: int = len(aws_resources_to_delete)
        num_resources_deleted: int = 0

        for resource in aws_resources_to_delete:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_deleted = resource.delete(aws_client=self.aws_api_client)
                if _resource_deleted:
                    num_resources_deleted += 1
                    print_info("Resource deleted")
                else:
                    logger.error(
                        f"Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                    )
                    if not self.aws_args.continue_on_delete_failure:
                        return False
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources deleted: {num_resources_deleted}/{num_resources_to_delete}"
        )
        if num_resources_to_delete == num_resources_deleted:
            return True

        logger.error(
            f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
        )
        return False

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Deleting AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return

        aws_resources_to_delete: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            sort_order="delete",
        )

        num_resources_to_delete: int = len(aws_resources_to_delete)
        print_heading("--**-- Aws resources:")
        for resource in aws_resources_to_delete:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_delete} resources")

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Patching AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return True

        aws_resources_to_patch: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        # Validate resources to be patched
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in aws_resources_to_patch:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            if self.aws_args.aws_region:
                print_info(f"\nRegion: {self.aws_args.aws_region}")
            if self.aws_args.aws_profile:
                print_info(f"Profile: {self.aws_args.aws_profile}")
            print_info(f"\nTotal {len(aws_resources_to_patch)} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping patch")
                print_info("-*-")
                exit(0)

        # track the total number of AwsResources to patch for validation
        num_resources_to_patch: int = len(aws_resources_to_patch)
        num_resources_patched: int = 0

        for resource in aws_resources_to_patch:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_patched = resource.update(aws_client=self.aws_api_client)
                if _resource_patched:
                    num_resources_patched += 1
                    print_info("Resource patched")
                else:
                    logger.error(
                        f"Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                    )
                    if not self.aws_args.continue_on_patch_failure:
                        return False
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources patched: {num_resources_patched}/{num_resources_to_patch}"
        )
        if num_resources_to_patch == num_resources_patched:
            return True

        logger.error(
            f"Resources patched: {num_resources_patched} do not match resources required: {num_resources_to_patch}"
        )
        return False

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Patching AwsResources")

        aws_resource_groups: Optional[
            Dict[str, AwsResourceGroup]
        ] = self.build_aws_resource_groups(app_filter=app_filter)

        if aws_resource_groups is None:
            print_info("No resources available")
            return

        aws_resources_to_patch: List[
            AwsResource
        ] = filter_and_flatten_aws_resource_groups(
            aws_resource_groups=aws_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        num_resources_to_patch: int = len(aws_resources_to_patch)
        print_heading("--**-- Aws resources:")
        for resource in aws_resources_to_patch:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        if self.aws_args.aws_region:
            print_info(f"\nRegion: {self.aws_args.aws_region}")
        if self.aws_args.aws_profile:
            print_info(f"Profile: {self.aws_args.aws_profile}")
        print_info(f"\nTotal {num_resources_to_patch} resources")

    ######################################################
    ## End
    ######################################################



from typing import Optional, List, Union

from phidata.app.base_app import BaseApp
from phidata.app.phidata_app import PhidataApp
from phidata.aws.resource.group import AwsResourceGroup
from phidata.infra.args import InfraArgs


class AwsArgs(InfraArgs):
    apps: Optional[List[Union[BaseApp, PhidataApp]]] = None

    # AwsResourceGroups to deploy
    resources: Optional[List[AwsResourceGroup]] = None
    # Resources dir where aws manifests are stored
    resources_dir: str = "aws"



from typing import Optional, List, Tuple, Dict, Any, Union
from typing_extensions import Literal

from phidata.asset.aws.aws_asset import AwsAsset, AwsAssetArgs
from phidata.aws.resource.s3.bucket import S3Bucket

from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class S3DatasetType(ExtendedEnum):
    CSV = "CSV"
    JSON = "JSON"
    PARQUET = "PARQUET"


class S3DatasetBaseArgs(AwsAssetArgs):
    # Dataset Name. Can be provided instead of "table" argument
    name: Optional[str] = None

    # Glue/Athena catalog: Table name.
    table: Optional[str] = None
    # Glue/Athena catalog: Database name.
    database: Optional[str] = None
    # Type of dataset: CSV, JSON, PARQUET
    dataset_type: Optional[S3DatasetType] = None
    # The type of the Glue Table. Set to EXTERNAL_TABLE if None.
    table_type: Optional[str] = None
    # Glue/Athena catalog: Table description
    table_description: Optional[str] = None
    # write mode – append (Default), overwrite, overwrite_partitions.
    write_mode: Optional[Literal["append", "overwrite", "overwrite_partitions"]] = None

    # S3 path for the dataset. If not provided, inferred from
    # path_prefix/top_level_dir/database/table or
    # s3://bucket/top_level_dir/database/table
    path: Optional[str] = None
    # Top level directory for all s3 datasets.
    top_level_dir: Optional[str] = "datasets"
    # Prefix for the dataset path. Used to infer dataset path if not provided as
    # path_prefix/top_level_dir/database/table
    path_prefix: Optional[str] = None
    # S3 bucket for the dataset. Used to infer dataset path if not provided as
    # s3://bucket/top_level_dir/database/table
    bucket: Optional[S3Bucket] = None
    # If dataset=True, add a filename prefix to the output files.
    filename_prefix: Optional[str] = None

    # Dictionary of columns names and Athena/Glue types to be casted.
    # Useful when you have columns with undetermined or mixed data types.
    # (e.g. {‘col name’: ‘bigint’, ‘col2 name’: ‘int’})
    dtype: Optional[Dict[str, str]] = None
    # Glue/Athena catalog: Key/value pairs to tag the table.
    parameters: Optional[Dict[str, str]] = None
    # Glue/Athena catalog: Columns names and the related comments
    # (e.g. {‘col0’: ‘Column 0.’, ‘col1’: ‘Column 1.’, ‘col2’: ‘Partition.’}).
    columns_comments: Optional[Dict[str, str]] = None

    # List of column names that will be used to create partitions.
    partition_cols: Optional[List[str]] = None
    # Tuple consisting of the column names used for
    # bucketing as the first element and the number of buckets as the second element.
    # Only str, int and bool are supported as column data types for bucketing.
    bucketing_info: Optional[Tuple[List[str], int]] = None
    # If True increases parallelism level during the partitions writing.
    # It will decrease the writing time and increase the memory usage.
    # https://aws-sdk-pandas.readthedocs.io/en/2.17.0/tutorials/022%20-%20Writing%20Partitions%20Concurrently.html
    concurrent_partitioning: Optional[bool] = None

    # If True and mode="overwrite", creates an archived version of the table catalog before updating it.
    catalog_versioning: Optional[bool] = None
    # If True allows schema evolution (new or missing columns), otherwise a exception will be raised.
    # True by default. (Only considered if mode in (“append”, “overwrite_partitions”))
    schema_evolution: Optional[bool] = None
    # The ID of the Data Catalog from which to retrieve Databases.
    # If none is provided, the AWS account ID is used by default.
    catalog_id: Optional[str] = None

    # True to enable concurrent requests, False to disable multiple threads.
    # If enabled os.cpu_count() will be used as the max number of threads.
    # If integer is provided, specified number is used.
    use_threads: Optional[Union[bool, int]] = None
    # Boto3 Session. The default boto3 Session will be used if boto3_session receive None.
    boto3_session: Optional[Any] = None
    # Forwarded to botocore requests.
    # e.g. s3_additional_kwargs={‘ServerSideEncryption’: ‘aws:kms’, ‘SSEKMSKeyId’: ‘YOUR_KMS_KEY_ARN’}
    s3_additional_kwargs: Optional[Dict[str, Any]] = None

    # Note: These arguments are not exposed using properties.
    # But they are included in args incase we want to pass down these values to aws wrangler.
    # regular_partitions (bool) – Create regular partitions (Non projected partitions) on Glue Catalog.
    # Disable when you will work only with Partition Projection.
    regular_partitions: Optional[bool] = None
    # Enable Partition Projection on Athena
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection.html
    projection_enabled: Optional[bool] = None
    # Dictionary of partitions names and Athena projections types. Valid types: “enum”, “integer”, “date”, “injected”
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘enum’, ‘col2_name’: ‘integer’}
    projection_types: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections ranges.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘0,10’, ‘col2_name’: ‘-1,8675309’}
    projection_ranges: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections values.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘A,B,Unknown’, ‘col2_name’: ‘foo,boo,bar’}
    projection_values: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections intervals.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘1’, ‘col2_name’: ‘5’}
    projection_intervals: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections digits.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘1’, ‘col2_name’: ‘2’}
    projection_digits: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections formats.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_date’: ‘yyyy-MM-dd’, ‘col2_timestamp’: ‘yyyy-MM-dd HH:mm:ss’}
    projection_formats: Optional[Dict[str, str]] = None
    # Value which is allows Athena to properly map partition values if the S3 file locations
    # do not follow a typical …/column=value/… pattern.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-setting-up.html
    # e.g. s3://bucket/table_root/a=${a}/${b}/some_static_subdirectory/${c}/
    projection_storage_location_template: Optional[str] = None


class S3DatasetBase(AwsAsset):
    def __init__(self) -> None:
        super().__init__()
        self.args: S3DatasetBaseArgs = S3DatasetBaseArgs()

    @property
    def table(self) -> Optional[str]:
        return self.args.table

    @table.setter
    def table(self, table: str) -> None:
        if table is not None:
            self.args.table = table

    @property
    def database(self) -> Optional[str]:
        return self.args.database

    @database.setter
    def database(self, database: str) -> None:
        if database is not None:
            self.args.database = database

    @property
    def dataset_type(self) -> Optional[S3DatasetType]:
        return self.args.dataset_type

    @dataset_type.setter
    def dataset_type(self, dataset_type: S3DatasetType) -> None:
        if dataset_type is not None:
            self.args.dataset_type = dataset_type

    @property
    def table_type(self) -> Optional[str]:
        return self.args.table_type

    @table_type.setter
    def table_type(self, table_type: str) -> None:
        if table_type is not None:
            self.args.table_type = table_type

    @property
    def table_description(self) -> Optional[str]:
        return self.args.table_description

    @table_description.setter
    def table_description(self, table_description: str) -> None:
        if table_description is not None:
            self.args.table_description = table_description

    @property
    def write_mode(
        self,
    ) -> Optional[Literal["append", "overwrite", "overwrite_partitions"]]:
        return self.args.write_mode

    @write_mode.setter
    def write_mode(
        self, write_mode: Literal["append", "overwrite", "overwrite_partitions"]
    ) -> None:
        if write_mode is not None:
            self.args.write_mode = write_mode

    @property
    def path(self) -> Optional[str]:
        return self.args.path

    @path.setter
    def path(self, path: str) -> None:
        if path is not None:
            self.args.path = path

    @property
    def top_level_dir(self) -> Optional[str]:
        return self.args.top_level_dir

    @top_level_dir.setter
    def top_level_dir(self, top_level_dir: str) -> None:
        if top_level_dir is not None:
            self.args.top_level_dir = top_level_dir

    @property
    def path_prefix(self) -> Optional[str]:
        return self.args.path_prefix

    @path_prefix.setter
    def path_prefix(self, path_prefix: str) -> None:
        if path_prefix is not None:
            self.args.path_prefix = path_prefix

    @property
    def bucket(self) -> Optional[S3Bucket]:
        return self.args.bucket

    @bucket.setter
    def bucket(self, bucket: S3Bucket) -> None:
        if bucket is not None:
            self.args.bucket = bucket

    @property
    def filename_prefix(self) -> Optional[str]:
        return self.args.filename_prefix

    @filename_prefix.setter
    def filename_prefix(self, filename_prefix: str) -> None:
        if filename_prefix is not None:
            self.args.filename_prefix = filename_prefix

    @property
    def dtype(self) -> Optional[Dict[str, str]]:
        return self.args.dtype

    @dtype.setter
    def dtype(self, dtype: Dict[str, str]) -> None:
        if dtype is not None:
            self.args.dtype = dtype

    @property
    def parameters(self) -> Optional[Dict[str, str]]:
        return self.args.parameters

    @parameters.setter
    def parameters(self, parameters: Dict[str, str]) -> None:
        if parameters is not None:
            self.args.parameters = parameters

    @property
    def columns_comments(self) -> Optional[Dict[str, str]]:
        return self.args.columns_comments

    @columns_comments.setter
    def columns_comments(self, columns_comments: Dict[str, str]) -> None:
        if columns_comments is not None:
            self.args.columns_comments = columns_comments

    @property
    def partition_cols(self) -> Optional[List[str]]:
        return self.args.partition_cols

    @partition_cols.setter
    def partition_cols(self, partition_cols: List[str]) -> None:
        if partition_cols is not None:
            self.args.partition_cols = partition_cols

    @property
    def bucketing_info(self) -> Optional[Tuple[List[str], int]]:
        return self.args.bucketing_info

    @bucketing_info.setter
    def bucketing_info(self, bucketing_info: Tuple[List[str], int]) -> None:
        if bucketing_info is not None:
            self.args.bucketing_info = bucketing_info

    @property
    def concurrent_partitioning(self) -> Optional[bool]:
        return self.args.concurrent_partitioning

    @concurrent_partitioning.setter
    def concurrent_partitioning(self, concurrent_partitioning: bool) -> None:
        if concurrent_partitioning is not None:
            self.args.concurrent_partitioning = concurrent_partitioning

    @property
    def catalog_versioning(self) -> Optional[bool]:
        return self.args.catalog_versioning

    @catalog_versioning.setter
    def catalog_versioning(self, catalog_versioning: bool) -> None:
        if catalog_versioning is not None:
            self.args.catalog_versioning = catalog_versioning

    @property
    def schema_evolution(self) -> Optional[bool]:
        return self.args.schema_evolution

    @schema_evolution.setter
    def schema_evolution(self, schema_evolution: bool) -> None:
        if schema_evolution is not None:
            self.args.schema_evolution = schema_evolution

    @property
    def catalog_id(self) -> Optional[str]:
        return self.args.catalog_id

    @catalog_id.setter
    def catalog_id(self, catalog_id: str) -> None:
        if catalog_id is not None:
            self.args.catalog_id = catalog_id

    @property
    def use_threads(self) -> Optional[Union[bool, int]]:
        return self.args.use_threads

    @use_threads.setter
    def use_threads(self, use_threads: Union[bool, int]) -> None:
        if use_threads is not None:
            self.args.use_threads = use_threads

    @property
    def boto3_session(self) -> Optional[Any]:
        return self.args.boto3_session

    @boto3_session.setter
    def boto3_session(self, boto3_session: Any) -> None:
        if boto3_session is not None:
            self.args.boto3_session = boto3_session

    @property
    def s3_additional_kwargs(self) -> Optional[Dict[str, Any]]:
        return self.args.s3_additional_kwargs

    @s3_additional_kwargs.setter
    def s3_additional_kwargs(self, s3_additional_kwargs: Dict[str, Any]) -> None:
        if s3_additional_kwargs is not None:
            self.args.s3_additional_kwargs = s3_additional_kwargs

    @property
    def uri(self) -> str:
        if self.path is not None:
            return self.path
        else:
            if self.path_prefix is not None:
                return "{}/{}{}/{}".format(
                    self.path_prefix,
                    f"{self.top_level_dir}/" if self.top_level_dir else "",
                    self.database,
                    self.table,
                )
            if self.bucket is not None:
                return "{}/{}{}/{}".format(
                    self.bucket.get_uri(),
                    f"{self.top_level_dir}/" if self.top_level_dir else "",
                    self.database,
                    self.table,
                )
        return ""



from typing import Optional, Dict, Any

from pydantic import BaseModel


class PhidataBaseArgs(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    enabled: bool = True

    #  -*- Resource Control
    skip_create: bool = False
    skip_read: bool = False
    skip_update: bool = False
    recreate_on_update: bool = False
    skip_delete: bool = False
    wait_for_creation: bool = True
    wait_for_update: bool = True
    wait_for_deletion: bool = True
    waiter_delay: int = 30
    waiter_max_attempts: int = 50

    #  -*- Save Resources to output directory
    # If True, save the resources to files
    save_output: bool = False
    # The resource directory for the output files
    resource_dir: Optional[str] = None

    # Skip creation if resource with the same name is active
    use_cache: bool = True

    class Config:
        arbitrary_types_allowed = True


class PhidataBase:
    """Phidata Base Class"""

    def __init__(self) -> None:
        self.args: PhidataBaseArgs = PhidataBaseArgs()

    @property
    def name(self) -> str:
        return self.args.name or self.__class__.__name__

    @property
    def version(self) -> Optional[str]:
        return self.args.version

    @property
    def enabled(self) -> bool:
        return self.args.enabled



from pathlib import Path
from typing import Any, Optional, Type, Union, Dict, List

from pydantic import BaseModel, validator

from phidata.utils.log import logger


class InfraResource(BaseModel):
    """Base class for all Phidata infrastructure resources.
    All Models in the phidata.*.resource modules are expected to be subclasses of this Model.

    We use a pydantic model for resources because the data which creates the resource
    may come from an external sources like users or an api.
    This data needs to be validated & type checked for which we use pydantic.
    """

    # name of resource
    name: Optional[str] = None
    # type of resource
    resource_type: Optional[str] = None

    # resource management
    # If True, skip resource creation if an active resources with the same name exists.
    use_cache: bool = True
    # Force the function to be implemented
    # i.e. force recreate, force delete, force update
    force: bool = False

    # If enabled=False: mark skip_create, skip_delete, skip_update as True
    enabled: bool = True
    # If True, phi does not create the resource
    skip_create: bool = False
    # If True, phi does not read the resource
    skip_read: bool = False
    # If True, phi does not update the resource
    skip_update: bool = False
    # If True, recreate the resource on update
    # Used for deployments with EBS volumes
    recreate_on_update: bool = False
    # If True, phi does not delete the resource
    skip_delete: bool = False

    # If True, waits for the resource to be created
    wait_for_creation: bool = True
    # If True, waits for the resource to be updated
    wait_for_update: bool = True
    # If True, waits for the resource to be deleted
    wait_for_deletion: bool = True
    # The amount of time in seconds to wait between attempts.
    waiter_delay: int = 30
    # The maximum number of attempts to be made.
    waiter_max_attempts: int = 50

    # Active resource object
    active_resource: Optional[Any] = None
    # Deprecated: the class of the active resource
    active_resource_class: Optional[Type] = None

    # If True, save the resource to a file
    save_output: bool = False
    resource_file: Optional[Union[str, Path]] = None
    # Add a resource directory to the resource file path
    resource_dir: Optional[str] = None

    # Other resources this resource depends on
    # Dependencies are always created if this resource is created
    depends_on: Optional[List[Any]] = None

    # Add secret variables to resource where applicable
    secret_data: Optional[Dict[str, Any]] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None
    # Add env variables to resource where applicable
    env_data: Optional[Dict[str, Any]] = None
    # Read env from a file in yaml format
    env_file: Optional[Path] = None

    def get_resource_name(self) -> Optional[str]:
        return self.name

    def get_resource_type(self) -> Optional[str]:
        return self.resource_type

    @validator("force", pre=True, always=True)
    def set_force(cls, force):
        from os import getenv

        force = force or getenv("PHI_WS_FORCE", False)
        return force

    @validator("skip_create", pre=True, always=True)
    def set_skip_create(cls, skip_create, values):
        skip_resource = not values.get("enabled", True)
        return True if skip_resource else skip_create

    @validator("skip_update", pre=True, always=True)
    def set_skip_update(cls, skip_update, values):
        skip_resource = not values.get("enabled", True)
        return True if skip_resource else skip_update

    @validator("skip_delete", pre=True, always=True)
    def set_skip_delete(cls, skip_delete, values):
        skip_resource = not values.get("enabled", True)
        return True if skip_resource else skip_delete

    """
    ## Functions to be implemented by subclasses
    def create(self, infra_api_client: InfraApiClient) -> bool:
    def read(self, infra_api_client: InfraApiClient) -> bool:
    def update(self, infra_api_client: InfraApiClient) -> bool:
    def delete(self, infra_api_client: InfraApiClient) -> bool:
    def is_active(self, infra_api_client: InfraApiClient) -> bool:
    """

    class Config:
        # https://pydantic-docs.helpmanual.io/usage/model_config/
        # If we need to use an alias for fields of subclasses, eg: Kubeconfig
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True

    @staticmethod
    def get_workspace_config_file_path() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKSPACE_CONFIG_FILE_ENV_VAR

        workspace_config_file_path = getenv(WORKSPACE_CONFIG_FILE_ENV_VAR, None)
        if workspace_config_file_path is not None:
            return Path(workspace_config_file_path)
        return None

    @staticmethod
    def get_workspace_root_path() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKSPACE_ROOT_ENV_VAR

        workspace_root_path = getenv(WORKSPACE_ROOT_ENV_VAR, None)
        if workspace_root_path is not None:
            return Path(workspace_root_path)
        return None

    @staticmethod
    def get_workspace_config_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKSPACE_CONFIG_DIR_ENV_VAR

        workspace_config_dir = getenv(WORKSPACE_CONFIG_DIR_ENV_VAR, None)
        if workspace_config_dir is not None:
            return Path(workspace_config_dir)
        return None

    @staticmethod
    def get_workflows_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKFLOWS_DIR_ENV_VAR

        workflows_dir = getenv(WORKFLOWS_DIR_ENV_VAR, None)
        if workflows_dir is not None:
            return Path(workflows_dir)
        return None

    @staticmethod
    def get_meta_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import META_DIR_ENV_VAR

        meta_dir = getenv(META_DIR_ENV_VAR, None)
        if meta_dir is not None:
            return Path(meta_dir)
        return None

    @staticmethod
    def get_notebooks_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import NOTEBOOKS_DIR_ENV_VAR

        notebooks_dir = getenv(NOTEBOOKS_DIR_ENV_VAR, None)
        if notebooks_dir is not None:
            return Path(notebooks_dir)
        return None

    @staticmethod
    def get_products_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import PRODUCTS_DIR_ENV_VAR

        products_dir = getenv(PRODUCTS_DIR_ENV_VAR, None)
        if products_dir is not None:
            return Path(products_dir)
        return None

    @staticmethod
    def get_scripts_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import SCRIPTS_DIR_ENV_VAR

        scripts_dir = getenv(SCRIPTS_DIR_ENV_VAR, None)
        if scripts_dir is not None:
            return Path(scripts_dir)
        return None

    @staticmethod
    def get_storage_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import STORAGE_DIR_ENV_VAR

        storage_dir = getenv(STORAGE_DIR_ENV_VAR, None)
        if storage_dir is not None:
            return Path(storage_dir)
        return None

    def read_yaml_file(self, file_path: Optional[Path]) -> Optional[Dict[str, Any]]:
        if file_path is not None and file_path.exists() and file_path.is_file():
            import yaml

            logger.debug(f"Reading {file_path}")
            data_from_file = yaml.safe_load(file_path.read_text())
            if data_from_file is not None and isinstance(data_from_file, dict):
                return data_from_file
            else:
                logger.error(f"Invalid file: {file_path}")
        return None

    def get_resource_file_path(self) -> Optional[Path]:
        if self.resource_file is None:
            workspace_config_dir = self.get_workspace_config_dir()
            if workspace_config_dir is not None:
                if self.name is not None and self.resource_type is not None:
                    file_name = f"{self.name}.json"
                    resource_dir = self.resource_dir or self.resource_type
                    return workspace_config_dir.joinpath(
                        "output", resource_dir, file_name
                    )
        if isinstance(self.resource_file, str):
            return Path(self.resource_file)
        elif isinstance(self.resource_file, Path):
            return self.resource_file
        return None

    def save_resource_file(self) -> bool:
        resource_file_path: Optional[Path] = self.get_resource_file_path()
        if resource_file_path is not None:
            try:
                from phidata.utils.json_io import write_json_file

                if not resource_file_path.exists():
                    resource_file_path.parent.mkdir(parents=True, exist_ok=True)
                    resource_file_path.touch(exist_ok=True)
                write_json_file(resource_file_path, self.active_resource)
                logger.info(f"Resource saved to: {str(resource_file_path)}")
                return True
            except Exception as e:
                logger.error(f"Could not write resource to file {e}")
        return False

    def read_resource_from_file(self) -> Optional[Dict[str, Any]]:
        resource_file_path: Optional[Path] = self.get_resource_file_path()
        if resource_file_path is not None:
            try:
                from phidata.utils.json_io import read_json_file

                if resource_file_path.exists() and resource_file_path.is_file():
                    data_from_file = read_json_file(resource_file_path)
                    if data_from_file is not None and isinstance(data_from_file, dict):
                        return data_from_file
                    else:
                        logger.warning(
                            f"Could not read {self.name} from {resource_file_path}"
                        )
            except Exception as e:
                logger.error(f"Could not read resource from file {e}")
        return None

    def delete_resource_file(self) -> bool:
        resource_file_path: Optional[Path] = self.get_resource_file_path()
        if resource_file_path is not None:
            try:
                if resource_file_path.exists() and resource_file_path.is_file():
                    resource_file_path.unlink()
                    logger.debug(f"Resource file deleted: {str(resource_file_path)}")
                    return True
            except Exception as e:
                logger.error(f"Could not delete resource file {e}")
        return False

    def attribute(self, name: str) -> Optional[Any]:
        resource_attributes = self.read_resource_from_file()
        if resource_attributes is not None:
            if name in resource_attributes:
                return resource_attributes[name]
            else:
                logger.warning(f"Resource attribute not found: {name}")
        return None

    def get_secret_data(self) -> Optional[Dict[str, str]]:
        if self.secret_data is not None:
            return self.secret_data

        if self.secrets_file is not None:
            self.secret_data = self.read_yaml_file(self.secrets_file)
        return self.secret_data

    def get_env_data(self) -> Optional[Dict[str, str]]:
        if self.env_data is not None:
            return self.env_data

        if self.env_file is not None:
            self.env_data = self.read_yaml_file(self.env_file)
        return self.env_data

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if other.resource_type == self.resource_type:
            return self.name == other.name
        return False



from pathlib import Path
from typing import Optional, Dict, Any, Union, List

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.types.context import ContainerPathContext
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class WorkspaceVolumeType(ExtendedEnum):
    HostPath = "HostPath"
    EmptyDir = "EmptyDir"


class AppVolumeType(ExtendedEnum):
    HostPath = "HostPath"
    EmptyDir = "EmptyDir"
    AwsEbs = "AwsEbs"
    AwsEfs = "AwsEfs"
    PersistentVolume = "PersistentVolume"


class BaseAppArgs(PhidataBaseArgs):
    # -*- Path parameters
    # Path to the workspace root directory
    workspace_root_path: Optional[Path] = None
    # Path to the workspace config file
    workspace_config_file_path: Optional[Path] = None
    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root dir
    #   to build paths depending on the environments (local, docker, k8s)
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    meta_dir: Optional[str] = None
    products_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    workflows_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None

    # -*- Image Configuration
    # Image can be provided as a DockerImage object
    image: Optional[Any] = None
    # OR as image_name:image_tag
    image_name: Optional[str] = None
    image_tag: Optional[str] = None
    entrypoint: Optional[Union[str, List[str]]] = None
    command: Optional[Union[str, List[str]]] = None

    # -*- Debug Mode
    debug_mode: bool = False

    # -*- Python Configuration
    # Install python dependencies using a requirements.txt file
    install_requirements: bool = False
    # Path to the requirements.txt file relative to the workspace_root
    requirements_file: str = "requirements.txt"
    # Set the PYTHONPATH env var
    set_python_path: bool = False
    # Manually provide the PYTHONPATH
    python_path: Optional[str] = None
    # Add paths to the PYTHONPATH env var
    # If python_path is provided, this value is ignored
    add_python_paths: Optional[List[str]] = None

    # -*- Container Environment
    # Add env variables to container
    env: Optional[Dict[str, Any]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Add secret variables to container
    secrets: Optional[Dict[str, Any]] = None
    # Read secret variables from a file in yaml format
    secrets_file: Optional[Path] = None
    # Read secret variables from AWS Secrets
    aws_secrets: Optional[Any] = None

    # -*- Container Ports
    # Open a container port if open_container_port=True
    open_container_port: bool = False
    # Port number on the container
    container_port: int = 80
    # Port name
    container_port_name: str = "http"
    # Host port to map to the container port
    container_host_port: int = 80

    # -*- Workspace Volume
    # Mount the workspace directory on the container
    mount_workspace: bool = False
    workspace_volume_name: Optional[str] = None
    workspace_volume_type: Optional[WorkspaceVolumeType] = None
    # Path to mount the workspace volume inside the container
    workspace_dir_container_path: str = "/mnt/workspaces"
    # Add the workspace name to the container path
    add_workspace_name_to_container_path: bool = True
    # -*- If workspace_volume_type=WorkspaceVolumeType.HostPath
    # Mount workspace_dir to workspace_dir_container_path
    # If None, use the workspace_root
    workspace_dir: Optional[str] = None
    # -*- If workspace_volume_type=WorkspaceVolumeType.EmptyDir
    # Then we can load the workspace from git using a git-sync sidecar
    create_git_sync_sidecar: bool = False
    # Required to create an initial copy of the workspace
    create_git_sync_init_container: bool = True
    git_sync_image_name: str = "k8s.gcr.io/git-sync"
    git_sync_image_tag: str = "v3.1.1"
    git_sync_repo: Optional[str] = None
    git_sync_branch: Optional[str] = None
    git_sync_wait: int = 1

    # -*- App Volume
    # Create a volume for mounting app data like notebooks, models, etc.
    create_app_volume: bool = False
    app_volume_name: Optional[str] = None
    app_volume_type: AppVolumeType = AppVolumeType.EmptyDir
    # Path to mount the app volume inside the container
    app_volume_container_path: str = "/mnt/app"
    # -*- If volume_type=AppVolumeType.HostPath
    app_volume_host_path: Optional[str] = None
    # -*- If volume_type=AppVolumeType.AwsEbs
    # EbsVolume: used to derive the volume_id, region, and az
    app_ebs_volume: Optional[Any] = None
    app_ebs_volume_region: Optional[str] = None
    app_ebs_volume_az: Optional[str] = None
    # Provide Ebs Volume-id manually
    app_ebs_volume_id: Optional[str] = None
    # -*- If volume_type=AppVolumeType.PersistentVolume
    # AccessModes is a list of ways the volume can be mounted.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#access-modes
    # Type: phidata.infra.k8s.enums.pv.PVAccessMode
    app_pv_access_modes: Optional[List[Any]] = None
    app_pv_requests_storage: Optional[str] = None
    # A list of mount options, e.g. ["ro", "soft"]. Not validated - mount will simply fail if one is invalid.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes/#mount-options
    app_pv_mount_options: Optional[List[str]] = None
    # What happens to a persistent volume when released from its claim.
    #   The default policy is Retain.
    # Literal["Delete", "Recycle", "Retain"]
    app_pv_reclaim_policy: Optional[str] = None
    app_pv_storage_class: str = ""
    app_pv_labels: Optional[Dict[str, str]] = None
    # -*- If volume_type=AppVolumeType.AwsEfs
    app_efs_volume_id: Optional[str] = None

    # Add NodeSelectors to Pods, so they are scheduled in the same region and zone as the ebs_volume
    schedule_pods_in_ebs_topology: bool = True

    # -*- AWS Configuration
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None

    #  -*- Other args
    print_env_on_load: bool = False

    # -*- Deprecated
    # Env vars added to docker env when building PhidataApps
    #   and running workflows
    docker_env: Optional[Dict[str, str]] = None
    # Env vars added to k8s env when building PhidataApps
    #   and running workflows
    k8s_env: Optional[Dict[str, str]] = None


class BaseApp(PhidataBase):
    """Base Class for all PhidataApps"""

    def __init__(self) -> None:
        super().__init__()

        # Cache env_data & secret_data
        self.env_data: Optional[Dict[str, Any]] = None
        self.secret_data: Optional[Dict[str, Any]] = None

        # Env variables used by the app
        self.container_env: Optional[Dict[str, str]] = None
        # Container paths used by the app
        self.container_paths: Optional[ContainerPathContext] = None

        # Args for the BaseApp, updated by the subclass
        self.args: BaseAppArgs = BaseAppArgs()

    @property
    def workspace_root_path(self) -> Optional[Path]:
        return self.args.workspace_root_path

    @workspace_root_path.setter
    def workspace_root_path(self, workspace_root_path: Path) -> None:
        if self.args is not None and workspace_root_path is not None:
            self.args.workspace_root_path = workspace_root_path

    @property
    def workspace_config_file_path(self) -> Optional[Path]:
        return self.args.workspace_config_file_path if self.args else None

    @workspace_config_file_path.setter
    def workspace_config_file_path(self, workspace_config_file_path: Path) -> None:
        if self.args is not None and workspace_config_file_path is not None:
            self.args.workspace_config_file_path = workspace_config_file_path

    @property
    def scripts_dir(self) -> Optional[str]:
        return self.args.scripts_dir

    @scripts_dir.setter
    def scripts_dir(self, scripts_dir: str) -> None:
        if self.args is not None and scripts_dir is not None:
            self.args.scripts_dir = scripts_dir

    @property
    def storage_dir(self) -> Optional[str]:
        return self.args.storage_dir

    @storage_dir.setter
    def storage_dir(self, storage_dir: str) -> None:
        if self.args is not None and storage_dir is not None:
            self.args.storage_dir = storage_dir

    @property
    def meta_dir(self) -> Optional[str]:
        return self.args.meta_dir

    @meta_dir.setter
    def meta_dir(self, meta_dir: str) -> None:
        if self.args is not None and meta_dir is not None:
            self.args.meta_dir = meta_dir

    @property
    def products_dir(self) -> Optional[str]:
        return self.args.products_dir

    @products_dir.setter
    def products_dir(self, products_dir: str) -> None:
        if self.args is not None and products_dir is not None:
            self.args.products_dir = products_dir

    @property
    def notebooks_dir(self) -> Optional[str]:
        return self.args.notebooks_dir

    @notebooks_dir.setter
    def notebooks_dir(self, notebooks_dir: str) -> None:
        if self.args is not None and notebooks_dir is not None:
            self.args.notebooks_dir = notebooks_dir

    @property
    def workflows_dir(self) -> Optional[str]:
        return self.args.workflows_dir

    @workflows_dir.setter
    def workflows_dir(self, workflows_dir: str) -> None:
        if self.args is not None and workflows_dir is not None:
            self.args.workflows_dir = workflows_dir

    @property
    def workspace_config_dir(self) -> Optional[str]:
        return self.args.workspace_config_dir

    @workspace_config_dir.setter
    def workspace_config_dir(self, workspace_config_dir: str) -> None:
        if self.args is not None and workspace_config_dir is not None:
            self.args.workspace_config_dir = workspace_config_dir

    @property
    def docker_env(self) -> Optional[Dict[str, str]]:
        return self.args.docker_env

    @docker_env.setter
    def docker_env(self, docker_env: Dict[str, str]) -> None:
        if self.args is not None and docker_env is not None:
            self.args.docker_env = docker_env

    @property
    def k8s_env(self) -> Optional[Dict[str, str]]:
        return self.args.k8s_env

    @k8s_env.setter
    def k8s_env(self, k8s_env: Dict[str, str]) -> None:
        if self.args is not None and k8s_env is not None:
            self.args.k8s_env = k8s_env

    @property
    def container_port(self) -> int:
        return self.args.container_port

    @container_port.setter
    def container_port(self, container_port: int) -> None:
        if self.args is not None and container_port is not None:
            self.args.container_port = container_port

    @property
    def container_host_port(self) -> int:
        return self.args.container_host_port

    @container_host_port.setter
    def container_host_port(self, container_host_port: int) -> None:
        if self.args is not None and container_host_port is not None:
            self.args.container_host_port = container_host_port

    @property
    def aws_region(self) -> Optional[str]:
        return self.args.aws_region

    @aws_region.setter
    def aws_region(self, aws_region: str) -> None:
        if self.args is not None and aws_region is not None:
            self.args.aws_region = aws_region

    @property
    def aws_profile(self) -> Optional[str]:
        return self.args.aws_profile

    @aws_profile.setter
    def aws_profile(self, aws_profile: str) -> None:
        if self.args is not None and aws_profile is not None:
            self.args.aws_profile = aws_profile

    @property
    def aws_config_file(self) -> Optional[str]:
        return self.args.aws_config_file

    @aws_config_file.setter
    def aws_config_file(self, aws_config_file: str) -> None:
        if self.args is not None and aws_config_file is not None:
            self.args.aws_config_file = aws_config_file

    @property
    def aws_shared_credentials_file(self) -> Optional[str]:
        return self.args.aws_shared_credentials_file

    @aws_shared_credentials_file.setter
    def aws_shared_credentials_file(self, aws_shared_credentials_file: str) -> None:
        if self.args is not None and aws_shared_credentials_file is not None:
            self.args.aws_shared_credentials_file = aws_shared_credentials_file

    def get_image_str(self):
        if self.args.image:
            return f"{self.args.image.name}:{self.args.image.tag}"
        elif self.args.image_name and self.args.image_tag:
            return f"{self.args.image_name}:{self.args.image_tag}"
        elif self.args.image_name:
            return f"{self.args.image_name}:latest"
        else:
            return None

    def get_env_data(self) -> Optional[Dict[str, str]]:
        if self.env_data is None:
            from phidata.utils.yaml_io import read_yaml_file

            self.env_data = read_yaml_file(file_path=self.args.env_file)
        return self.env_data

    def get_secret_data(self) -> Optional[Dict[str, str]]:
        if self.secret_data is None:
            from phidata.utils.yaml_io import read_yaml_file

            self.secret_data = read_yaml_file(file_path=self.args.secrets_file)
        return self.secret_data

    def get_container_paths(self) -> Optional[ContainerPathContext]:
        if self.container_paths is not None:
            return self.container_paths

        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None

        workspace_name = self.workspace_root_path.stem
        if workspace_name is None:
            return None

        workspace_volume_container_path: str = self.args.workspace_dir_container_path
        if workspace_volume_container_path is None:
            return None

        workspace_root_container_path = (
            f"{workspace_volume_container_path}/{workspace_name}"
            if self.args.add_workspace_name_to_container_path
            else workspace_volume_container_path
        )
        container_paths = ContainerPathContext(
            workspace_name=workspace_name,
            workspace_root=workspace_root_container_path,
            # Required for git-sync and K8s volume mounts
            workspace_parent=workspace_volume_container_path,
        )

        if self.args.scripts_dir is not None:
            container_paths.scripts_dir = (
                f"{workspace_root_container_path}/{self.args.scripts_dir}"
            )

        if self.args.storage_dir is not None:
            container_paths.storage_dir = (
                f"{workspace_root_container_path}/{self.args.storage_dir}"
            )

        if self.args.meta_dir is not None:
            container_paths.meta_dir = (
                f"{workspace_root_container_path}/{self.args.meta_dir}"
            )

        if self.args.products_dir is not None:
            container_paths.products_dir = (
                f"{workspace_root_container_path}/{self.args.products_dir}"
            )

        if self.args.notebooks_dir is not None:
            container_paths.notebooks_dir = (
                f"{workspace_root_container_path}/{self.args.notebooks_dir}"
            )

        if self.args.workflows_dir is not None:
            container_paths.workflows_dir = (
                f"{workspace_root_container_path}/{self.args.workflows_dir}"
            )

        if self.args.workspace_config_dir is not None:
            container_paths.workspace_config_dir = (
                f"{workspace_root_container_path}/{self.args.workspace_config_dir}"
            )

        if self.args.requirements_file is not None:
            container_paths.requirements_file = (
                f"{workspace_root_container_path}/{self.args.requirements_file}"
            )

        return container_paths

    def set_aws_env_vars(self, env_dict: Dict[str, str]) -> None:
        from phidata.constants import (
            AWS_REGION_ENV_VAR,
            AWS_DEFAULT_REGION_ENV_VAR,
        )

        if self.aws_region is not None:
            env_dict[AWS_REGION_ENV_VAR] = self.aws_region
            env_dict[AWS_DEFAULT_REGION_ENV_VAR] = self.aws_region

    def get_docker_resource_groups(
        self, docker_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Dict[str, Any]]:
        logger.debug(f"@get_docker_resource_groups not defined for {self.name}")
        return None

    def get_k8s_resource_groups(
        self, k8s_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Dict[str, Any]]:
        logger.debug(f"@get_k8s_resource_groups not defined for {self.name}")
        return None

    def get_aws_resource_groups(
        self, aws_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Dict[str, Any]]:
        logger.debug(f"@get_aws_resource_groups not defined for {self.name}")
        return None



from phidata.utils.log import logger

try:
    import duckdb
except ImportError:
    logger.warning("DuckDB not available.")
    raise


def create_duckdb_connection(
    db_path: str = ":memory:", s3_region: str = "us-east-1"
) -> duckdb.DuckDBPyConnection:
    """
    Create a duckdb connection

    Args:
        db_path (str, optional): Path to the database. Defaults to ":memory:".
        s3_region (str, optional): S3 region. Defaults to "us-east-1".

    Returns:
        duckdb.DuckDBPyConnection: duckdb connection
    """
    duckdb_connection = duckdb.connect(db_path)
    try:
        duckdb_connection.sql("INSTALL httpfs;")
        duckdb_connection.sql("LOAD httpfs;")
        duckdb_connection.sql(f"SET s3_region='{s3_region}';")
    except Exception:
        logger.warning(
            "Failed to install httpfs extension. Only local files will be supported"
        )

    return duckdb_connection



from phidata.app.amundsen.frontend import (
    AmundsenFrontend,
    AmundsenFrontendArgs,
    RestartPolicy,
    ImagePullPolicy,
)

from phidata.app.amundsen.metadata import (
    AmundsenMetadata,
    AmundsenMetadataArgs,
    RestartPolicy,
    ImagePullPolicy,
)

from phidata.app.amundsen.search import (
    AmundsenSearch,
    AmundsenSearchArgs,
    RestartPolicy,
    ImagePullPolicy,
)



from phidata.utils.log import logger

try:
    from langchain import LLMChain
    from langchain.agents import AgentExecutor, Tool, ZeroShotAgent
except ImportError:
    logger.warning("Langchain not available.")
    raise

from phidata.llm.duckdb.tool import DuckDBQueryTool
from phidata.llm.duckdb.query import run_duckdb_query, describe_table_or_view


def get_duckdb_query_chain(llm, duckdb_connection):
    tools = [
        Tool(
            name="Show Tables",
            func=lambda _: run_duckdb_query(duckdb_connection, "show tables;"),
            description="Useful to show the available tables and views. Empty input required.",
        ),
        Tool(
            name="Describe Table",
            func=lambda table: describe_table_or_view(duckdb_connection, table),
            description="Useful to show the column names and types of a table or view. Use a valid table name as the input.",  # noqa: E501
        ),
        Tool(
            name="Query Inspector",
            func=lambda query: query.strip('"').strip("'"),
            description="Useful to show the query before execution. Always inspect your query before execution. Input MUST be on one line.",  # noqa: E501
        ),
        DuckDBQueryTool(duckdb_connection=duckdb_connection),
    ]

    # https://langchain.readthedocs.io/en/latest/modules/agents/examples/custom_agent.html#custom-llmchain
    prompt = ZeroShotAgent.create_prompt(
        tools,
        prefix=prefix,
        suffix=suffix,
        input_variables=["input", "agent_scratchpad", "table_names"],
    )

    llm_chain = LLMChain(llm=llm, prompt=prompt)

    tool_names = [tool.name for tool in tools]

    agent = ZeroShotAgent(
        llm_chain=llm_chain,
        allowed_tools=tool_names,
    )
    agent_executor = AgentExecutor.from_agent_and_tools(
        agent=agent,
        tools=tools,
        verbose=True,
    )

    return agent_executor


prefix = """Given an input question, identify the relevant tables and relevant columns, then create
one single syntactically correct DuckDB query to inspect, then execute, before returning the answer.
If the input is a valid looking SQL query selecting data or creating a view, execute it directly.

Even if you know the answer, you MUST show you can get the answer from the database.
Inspect your query before execution.

Always share the SQL queries you use to get the answer.

Refuse to delete any data, or drop tables. You only execute one statement at a time. You may import data.

Example imports:
- CREATE table customers AS SELECT * FROM 'data/records.json';
- CREATE VIEW covid AS SELECT * FROM 's3://covid19-lake/data.csv';

Unless the user specifies in their question a specific number of examples to obtain, limit your
query to at most 5 results. You can order the results by a relevant column to return the most interesting
examples in the database.

Pay attention to use only the column names that you can see in the schema description. Pay attention
to which column is in which table.

You have access to the following tables/views:
{table_names}

You have access to the following tools:
"""

suffix = """After outputting the Action Input you never output an Observation, that will be provided to you.

Always ist the relevant SQL queries you ran in your final answer.

If a query fails, try fix it, if the database doesn't contain the answer, or returns no results,
output a summary of your actions in your final answer. It is important that you use the exact format:

Final Answer: I have successfully created a view of the data.

Queries should be output on one line and don't use any escape characters.

Let's go! Remember it is important that you use the exact phrase "Final Answer: " to begin your
final answer.

Question: {input}
Thought: I should describe the most relevant tables in the database to see what columns will be useful.
{agent_scratchpad}"""



from pathlib import Path
from typing import Optional, Any, List, Dict
from typing_extensions import Literal

from phidata.asset.local import LocalAsset, LocalAssetArgs
from phidata.checks.check import Check
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class LocalTableFormat(ExtendedEnum):
    CSV = "csv"
    IPC = "ipc"
    ARROW = "arrow"
    FEATHER = "feather"
    ORC = "orc"
    PARQUET = "parquet"


class LocalTableArgs(LocalAssetArgs):
    # Table Name
    name: str
    # Table Format
    table_format: LocalTableFormat
    # Database for the table
    database: str

    # -*- Table Path
    # If current_dir=True, store the table in the current directory
    current_dir: bool = False
    # Top level directory for all tables, under the "storage" directory
    top_level_dir: Optional[str] = "tables"
    # Provide absolute path to the table directory
    table_dir: Optional[str] = None
    # A template string used to generate basenames of written data files.
    # The token ‘{i}’ will be replaced with an automatically incremented integer.
    # If not specified, it defaults to “part-{i}.” + format.default_extname
    basename_template: Optional[str] = None

    # List of partition columns
    partitions: Optional[List[str]] = None

    # Maximum number of partitions any batch may be written into.
    max_partitions: Optional[int] = None
    # If greater than 0 then this will limit the maximum number of files that can be left open.
    # If an attempt is made to open too many files then the least recently used file will be closed.
    # If this setting is set too low you may end up fragmenting your data into many small files.
    max_open_files: Optional[int] = None
    # Maximum number of rows per file. If greater than 0 then this will limit how many rows are placed in any single
    # file. Otherwise there will be no limit and one file will be created in each output directory unless files need
    # to be closed to respect max_open_files
    max_rows_per_file: Optional[int] = None
    # Minimum number of rows per group. When the value is greater than 0, the dataset writer will batch incoming data
    # and only write the row groups to the disk when sufficient rows have accumulated.
    min_rows_per_group: Optional[int] = None
    # Maximum number of rows per group. If the value is greater than 0, then the dataset writer may split up large
    # incoming batches into multiple row groups. If this value is set, then min_rows_per_group should also be set.
    # Otherwise it could end up with very small row groups.
    max_rows_per_group: Optional[int] = None
    # Controls how the dataset will handle data that already exists in the destination.
    # The default behavior (‘error’) is to raise an error if any data exists in the destination.
    # ‘overwrite_or_ignore’ will ignore any existing data and will overwrite files with the same name
    # as an output file. Other existing files will be ignored. This behavior, in combination with a unique
    # basename_template for each write, will allow for an append workflow.
    # ‘delete_matching’ is useful when you are writing a partitioned dataset.
    # The first time each partition directory is encountered the entire directory will be deleted.
    # This allows you to overwrite old partitions completely.
    write_mode: Literal[
        "delete_matching", "overwrite_or_ignore", "error"
    ] = "delete_matching"


class LocalTable(LocalAsset):
    """Base Class for Local tables"""

    def __init__(
        self,
        # Table Name: required
        name: str,
        # Table Format: required
        table_format: LocalTableFormat,
        # Database for the table
        database: str = "default",
        # DataModel for this table
        data_model: Optional[Any] = None,
        # Checks to run before reading from disk
        read_checks: Optional[List[Check]] = None,
        # Checks to run before writing to disk
        write_checks: Optional[List[Check]] = None,
        # -*- Table Path
        # If current_dir=True, store the table in the current directory
        current_dir: bool = False,
        # Top level directory for all tables, under the "storage" directory
        top_level_dir: Optional[str] = "tables",
        # Provide absolute path to the table directory
        table_dir: Optional[str] = None,
        # A template string used to generate basenames of written data files.
        # The token ‘{i}’ will be replaced with an automatically incremented integer.
        # If not specified, it defaults to “part-{i}.” + format.default_extname
        basename_template: Optional[str] = None,
        # List of partition columns
        partitions: Optional[List[str]] = None,
        # Maximum number of partitions any batch may be written into.
        max_partitions: Optional[int] = None,
        # If greater than 0 then this will limit the maximum number of files that can be left open.
        # If an attempt is made to open too many files then the least recently used file will be closed.
        # If this setting is set too low you may end up fragmenting your data into many small files.
        max_open_files: Optional[int] = None,
        # Maximum number of rows per file. If greater than 0 then this will limit how many rows are placed in any single
        # file. Otherwise there will be no limit and one file will be created in each output directory unless files need
        # to be closed to respect max_open_files
        max_rows_per_file: Optional[int] = None,
        # Minimum number of rows per group. When the value is greater than 0, the dataset writer will batch incoming data
        # and only write the row groups to the disk when sufficient rows have accumulated.
        min_rows_per_group: Optional[int] = None,
        # Maximum number of rows per group. If the value is greater than 0, then the dataset writer may split up large
        # incoming batches into multiple row groups. If this value is set, then min_rows_per_group should also be set.
        # Otherwise it could end up with very small row groups.
        max_rows_per_group: Optional[int] = None,
        # Controls how the dataset will handle data that already exists in the destination.
        # The default behavior (‘error’) is to raise an error if any data exists in the destination.
        # ‘overwrite_or_ignore’ will ignore any existing data and will overwrite files with the same name
        # as an output file. Other existing files will be ignored. This behavior, in combination with a unique
        # basename_template for each write, will allow for an append workflow.
        # ‘delete_matching’ is useful when you are writing a partitioned dataset.
        # The first time each partition directory is encountered the entire directory will be deleted.
        # This allows you to overwrite old partitions completely.
        write_mode: Literal[
            "delete_matching", "overwrite_or_ignore", "error"
        ] = "delete_matching",
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()
        try:
            self.args: LocalTableArgs = LocalTableArgs(
                name=name,
                table_format=table_format,
                database=database,
                data_model=data_model,
                read_checks=read_checks,
                write_checks=write_checks,
                current_dir=current_dir,
                top_level_dir=top_level_dir,
                table_dir=table_dir,
                basename_template=basename_template,
                partitions=partitions,
                max_partitions=max_partitions,
                max_open_files=max_open_files,
                max_rows_per_file=max_rows_per_file,
                min_rows_per_group=min_rows_per_group,
                max_rows_per_group=max_rows_per_group,
                write_mode=write_mode,
                version=version,
                enabled=enabled,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid")
            raise

    @property
    def table_format(self) -> Optional[LocalTableFormat]:
        return self.args.table_format if self.args else None

    @table_format.setter
    def table_format(self, table_format: LocalTableFormat) -> None:
        if self.args and table_format:
            self.args.table_format = table_format

    @property
    def database(self) -> Optional[str]:
        return self.args.database if self.args else None

    @database.setter
    def database(self, database: str) -> None:
        if self.args and database:
            self.args.database = database

    @property
    def current_dir(self) -> Optional[bool]:
        return self.args.current_dir if self.args else None

    @current_dir.setter
    def current_dir(self, current_dir: bool) -> None:
        if self.args and current_dir:
            self.args.current_dir = current_dir

    @property
    def top_level_dir(self) -> Optional[str]:
        return self.args.top_level_dir if self.args else None

    @top_level_dir.setter
    def top_level_dir(self, top_level_dir: str) -> None:
        if self.args and top_level_dir:
            self.args.top_level_dir = top_level_dir

    @property
    def table_dir(self) -> Optional[str]:
        return self.args.table_dir if self.args else None

    @table_dir.setter
    def table_dir(self, table_dir: str) -> None:
        if self.args and table_dir:
            self.args.table_dir = table_dir

    @property
    def basename_template(self) -> Optional[str]:
        return self.args.basename_template if self.args else None

    @basename_template.setter
    def basename_template(self, basename_template: str) -> None:
        if self.args and basename_template:
            self.args.basename_template = basename_template

    @property
    def partitions(self) -> Optional[List[str]]:
        return self.args.partitions if self.args else None

    @partitions.setter
    def partitions(self, partitions: List[str]) -> None:
        if self.args and partitions:
            self.args.partitions = partitions

    @property
    def max_partitions(self) -> Optional[int]:
        return self.args.max_partitions if self.args else None

    @max_partitions.setter
    def max_partitions(self, max_partitions: int) -> None:
        if self.args and max_partitions:
            self.args.max_partitions = max_partitions

    @property
    def max_open_files(self) -> Optional[int]:
        return self.args.max_open_files if self.args else None

    @max_open_files.setter
    def max_open_files(self, max_open_files: int) -> None:
        if self.args and max_open_files:
            self.args.max_open_files = max_open_files

    @property
    def max_rows_per_file(self) -> Optional[int]:
        return self.args.max_rows_per_file if self.args else None

    @max_rows_per_file.setter
    def max_rows_per_file(self, max_rows_per_file: int) -> None:
        if self.args and max_rows_per_file:
            self.args.max_rows_per_file = max_rows_per_file

    @property
    def min_rows_per_group(self) -> Optional[int]:
        return self.args.min_rows_per_group if self.args else None

    @min_rows_per_group.setter
    def min_rows_per_group(self, min_rows_per_group: int) -> None:
        if self.args and min_rows_per_group:
            self.args.min_rows_per_group = min_rows_per_group

    @property
    def max_rows_per_group(self) -> Optional[int]:
        return self.args.max_rows_per_group if self.args else None

    @max_rows_per_group.setter
    def max_rows_per_group(self, max_rows_per_group: int) -> None:
        if self.args and max_rows_per_group:
            self.args.max_rows_per_group = max_rows_per_group

    @property
    def write_mode(
        self,
    ) -> Optional[Literal["delete_matching", "overwrite_or_ignore", "error"]]:
        return self.args.write_mode if self.args else None

    @write_mode.setter
    def write_mode(
        self, write_mode: Literal["delete_matching", "overwrite_or_ignore", "error"]
    ) -> None:
        if self.args and write_mode:
            self.args.write_mode = write_mode

    @property
    def table_location(self) -> str:
        if self.table_dir is not None:
            return self.table_dir

        # logger.debug("-*- Building local table location")
        base_dir: Optional[Path] = None

        # Use current_dir as base path if set
        if self.current_dir:
            base_dir = Path(__file__).resolve()

        # Or use storage_dir_path as the base path
        if base_dir is None:
            # storage_dir_path is loaded from the current environment variable
            base_dir = self.storage_dir_path

        # Add the file_dir if provided
        if self.top_level_dir is not None:
            if base_dir is None:
                # base_dir is None meaning no storage_dir_path
                base_dir = Path(".").resolve().joinpath(self.top_level_dir)
            else:
                base_dir = base_dir.joinpath(self.top_level_dir)

        # Add the file_name
        if self.name is not None:
            if base_dir is None:
                base_dir = Path(".").resolve().joinpath(self.name)
            else:
                base_dir = base_dir.joinpath(self.name)

        self.table_dir = str(base_dir)
        logger.debug(f"-*- Table location: {self.table_dir}")
        return self.table_dir

    ######################################################
    ## Validate data asset
    ######################################################

    def is_valid(self) -> bool:
        return True

    ######################################################
    ## Build data asset
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.name}")
        return False

    ######################################################
    ## Write DataAsset
    ######################################################

    def write_table(self, table: Any, **write_options) -> bool:
        """
        Write pyarrow.Table to disk.

        https://arrow.apache.org/docs/python/generated/pyarrow.Table.html#pyarrow.Table
        """
        # Validate pyarrow is installed
        try:
            import pyarrow as pa
            import pyarrow.dataset as ds
        except ImportError as ie:
            raise Exception(
                f"PyArrow not installed. Please install with `pip install pyarrow`"
            ) from ie

        # Validate table
        if table is None or not isinstance(table, pa.Table):
            logger.error("Table invalid")
            return False

        # Check table_location is available
        table_location = self.table_location
        if table_location is None:
            logger.error("Table location invalid")
            return False

        # Check LocalFileSystem is available
        fs = self._get_fs()
        if fs is None:
            logger.error("Could not create LocalFileSystem")
            return False

        logger.debug("Format: {}".format(self.args.table_format.value))
        try:
            # Run write checks
            if self.write_checks is not None:
                for check in self.write_checks:
                    if not check.check_table(table):
                        return False

            # Create a dict of args which are not null
            not_null_args: Dict[str, Any] = {}
            if self.args.basename_template is not None:
                not_null_args["basename_template"] = self.args.basename_template
            if self.args.partitions is not None:
                not_null_args["partitioning"] = self.args.partitions
                not_null_args["partitioning_flavor"] = "hive"
                # cast partition keys to string
                # ref: https://bneijt.nl/blog/write-polars-dataframe-as-parquet-dataset/
                table = table.cast(
                    pa.schema(
                        [
                            f.with_type(pa.string())
                            if f.name in self.args.partitions
                            else f
                            for f in table.schema
                        ]
                    )
                )
            if self.args.max_partitions is not None:
                not_null_args["max_partitions"] = self.args.max_partitions
            if self.args.max_open_files is not None:
                not_null_args["max_open_files"] = self.args.max_open_files
            if self.args.max_rows_per_file is not None:
                not_null_args["max_rows_per_file"] = self.args.max_rows_per_file
            if self.args.min_rows_per_group is not None:
                not_null_args["min_rows_per_group"] = self.args.min_rows_per_group
            if self.args.max_rows_per_group is not None:
                not_null_args["max_rows_per_group"] = self.args.max_rows_per_group

            # Build file_options: FileFormat specific write options
            # created using the FileFormat.make_write_options() function.
            if write_options:
                file_options = ds.FileFormat.make_write_options(**write_options)
                not_null_args["file_options"] = file_options

            # Write table to disk
            ds.write_dataset(
                table,
                table_location,
                format=self.args.table_format.value,
                filesystem=fs,
                existing_data_behavior=self.args.write_mode,
                **not_null_args,
            )
            logger.info(f"Table {self.name} written to {table_location}")

            return True
        except Exception:
            logger.error("Could not write table: {}".format(self.name))
            raise

    def write_polars_df(
        self,
        df: Optional[Any] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> bool:
        """
        Write Polars DataFrame to disk.
        """
        # Validate polars is installed
        try:
            import polars as pl  # type: ignore
        except ImportError as ie:
            logger.error(f"Polars not installed: {ie}")
            return False

        # Validate df
        if df is None or not isinstance(df, pl.DataFrame):
            logger.error("DataFrame invalid")
            return False

        # Create arrow table and write to disk
        return self.write_table(df.to_arrow(), options=options)

    def write_pandas_df(self, df: Optional[Any] = None, **kwargs) -> bool:
        logger.debug(f"@write_pandas_df not defined for {self.name}")
        return False

    ######################################################
    ## Create DataAsset
    ######################################################

    def _create(self) -> bool:
        logger.error(f"@_create not defined for {self.name}")
        return False

    def post_create(self) -> bool:
        return True

    ######################################################
    ## Read DataAsset
    ######################################################

    def read_table(self, **read_options) -> Optional[Any]:
        """
        Read pyarrow.Table from disk.

        https://arrow.apache.org/docs/python/generated/pyarrow.Table.html#pyarrow.Table
        """
        # Validate pyarrow is installed
        try:
            import pyarrow as pa
            import pyarrow.dataset as ds
        except ImportError as ie:
            raise Exception(
                f"PyArrow not installed. Please install with `pip install pyarrow`"
            ) from ie

        # Check table_location is available
        table_location = self.table_location
        if table_location is None:
            logger.error("Table location invalid")
            return False

        # Check LocalFileSystem is available
        fs = self._get_fs()
        if fs is None:
            logger.error("Could not create LocalFileSystem")
            return False

        logger.debug("Format: {}".format(self.args.table_format.value))
        try:
            # Create a dict of args which are not null
            not_null_args: Dict[str, Any] = {}
            if self.args.partitions is not None:
                not_null_args["partitioning"] = "hive"

            # Read dataset from s3
            # https://arrow.apache.org/docs/python/generated/pyarrow.dataset.dataset.html#pyarrow.dataset.dataset
            # https://arrow.apache.org/docs/python/generated/pyarrow.dataset.Dataset.html#pyarrow.dataset.Dataset
            dataset: ds.Dataset = ds.dataset(
                table_location,
                format=self.args.table_format.value,
                filesystem=fs,
                **not_null_args,
            )

            # Convert dataset to table
            # https://arrow.apache.org/docs/python/generated/pyarrow.dataset.Dataset.html#pyarrow.dataset.Dataset.to_table
            # for available options, see:
            # https://arrow.apache.org/docs/python/generated/pyarrow.dataset.Scanner.html#pyarrow.dataset.Scanner.from_dataset
            table: pa.Table = dataset.to_table(**read_options)

            # Run read checks
            if self.read_checks is not None:
                for check in self.read_checks:
                    if not check.check_table(table):
                        return None

            return table
        except Exception:
            logger.error("Could not read table: {}".format(self.name))
            raise

    def read_polars_df(
        self, options: Optional[Dict[str, Any]] = None, **kwargs
    ) -> Optional[Any]:
        """
        Read Polars DataFrame from disk.

        options: Dict[str, Any]
            Additional options to pass to the reader.
            More info:
                https://arrow.apache.org/docs/python/generated/pyarrow.dataset.Scanner.html#pyarrow.dataset.Scanner.from_dataset
                https://arrow.apache.org/docs/python/generated/pyarrow.dataset.dataset.html#pyarrow.dataset.dataset.to_table
        """
        # Validate polars is installed
        try:
            import polars as pl
        except ImportError as ie:
            logger.error(f"Polars not installed: {ie}")
            return None

        # Convert table to polars DataFrame
        # https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.from_arrow.html
        return pl.from_arrow(self.read_table(options=options))

    def read_pandas_df(self, **kwargs) -> Optional[Any]:
        logger.debug(f"@read_pandas_df not defined for {self.name}")
        return False

    def _read(self) -> Any:
        logger.error(f"@_read not defined for {self.name}")
        return False

    ######################################################
    ## Update DataAsset
    ######################################################

    def _update(self) -> Any:
        logger.error(f"@_update not defined for {self.name}")
        return False

    def post_update(self) -> bool:
        return True

    ######################################################
    ## Delete DataAsset
    ######################################################

    def _delete(self) -> Any:
        logger.error(f"@_delete not defined for {self.name}")
        return False

    def post_delete(self) -> bool:
        return True



from typing import Optional, Dict, Any, Union, List
from collections import OrderedDict

from phidata.app.base_app import (
    BaseApp,
    BaseAppArgs,
    WorkspaceVolumeType,
    AppVolumeType,
)
from phidata.k8s.enums.restart_policy import RestartPolicy
from phidata.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.k8s.enums.service_type import ServiceType
from phidata.types.context import ContainerPathContext
from phidata.utils.log import logger


class K8sAppArgs(BaseAppArgs):
    # -*- Container Configuration
    container_name: Optional[str] = None
    container_labels: Optional[Dict[str, str]] = None

    # -*- Pod Configuration
    pod_name: Optional[str] = None
    pod_annotations: Optional[Dict[str, str]] = None
    pod_node_selector: Optional[Dict[str, str]] = None

    # -*- Secret Configuration
    secret_name: Optional[str] = None

    # -*- Configmap Configuration
    configmap_name: Optional[str] = None

    # -*- Deployment Configuration
    replicas: int = 1
    deploy_name: Optional[str] = None
    image_pull_policy: Optional[ImagePullPolicy] = None
    deploy_restart_policy: Optional[RestartPolicy] = None
    deploy_labels: Optional[Dict[str, Any]] = None
    termination_grace_period_seconds: Optional[int] = None
    # Key to spread the pods across a topology
    topology_spread_key: Optional[str] = None
    # The degree to which pods may be unevenly distributed
    topology_spread_max_skew: Optional[int] = None
    # How to deal with a pod if it doesn't satisfy the spread constraint.
    topology_spread_when_unsatisfiable: Optional[str] = None

    # -*- Service Configuration
    create_service: bool = False
    service_name: Optional[str] = None
    service_type: Optional[ServiceType] = None
    # The port exposed by the service.
    service_port: int = 8000
    # The node_port exposed by the service if service_type = ServiceType.NODE_PORT
    service_node_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    service_target_port: Optional[Union[str, int]] = None
    # Extra ports exposed by the webserver service. Type: List[CreatePort]
    service_ports: Optional[List[Any]] = None
    service_labels: Optional[Dict[str, Any]] = None
    service_annotations: Optional[Dict[str, str]] = None
    # If ServiceType == ServiceType.LoadBalancer
    service_health_check_node_port: Optional[int] = None
    service_internal_traffic_policy: Optional[str] = None
    service_load_balancer_class: Optional[str] = None
    service_load_balancer_ip: Optional[str] = None
    service_load_balancer_source_ranges: Optional[List[str]] = None
    service_allocate_load_balancer_node_ports: Optional[bool] = None

    # -*- Ingress Configuration
    create_ingress: bool = False
    ingress_name: Optional[str] = None
    ingress_annotations: Optional[Dict[str, str]] = None

    # -*- RBAC Configuration
    use_rbac: bool = False
    # Create a Namespace with name ns_name & default values
    ns_name: Optional[str] = None
    # or Provide the full Namespace definition
    # Type: CreateNamespace
    namespace: Optional[Any] = None
    # Create a ServiceAccount with name sa_name & default values
    sa_name: Optional[str] = None
    # or Provide the full ServiceAccount definition
    # Type: CreateServiceAccount
    service_account: Optional[Any] = None
    # Create a ClusterRole with name cr_name & default values
    cr_name: Optional[str] = None
    # or Provide the full ClusterRole definition
    # Type: CreateClusterRole
    cluster_role: Optional[Any] = None
    # Create a ClusterRoleBinding with name crb_name & default values
    crb_name: Optional[str] = None
    # or Provide the full ClusterRoleBinding definition
    # Type: CreateClusterRoleBinding
    cluster_role_binding: Optional[Any] = None

    # Add additional Kubernetes resources to the App
    # Type: CreateSecret
    extra_secrets: Optional[List[Any]] = None
    # Type: CreateConfigMap
    extra_configmaps: Optional[List[Any]] = None
    # Type: CreateService
    extra_services: Optional[List[Any]] = None
    # Type: CreateDeployment
    extra_deployments: Optional[List[Any]] = None
    # Type: CreatePersistentVolume
    extra_pvs: Optional[List[Any]] = None
    # Type: CreatePVC
    extra_pvcs: Optional[List[Any]] = None
    # Type: CreateContainer
    extra_containers: Optional[List[Any]] = None
    # Type: CreateContainer
    extra_init_containers: Optional[List[Any]] = None
    # Type: CreatePort
    extra_ports: Optional[List[Any]] = None
    # Type: CreateVolume
    extra_volumes: Optional[List[Any]] = None
    # Type: CreateStorageClass
    extra_storage_classes: Optional[List[Any]] = None
    # Type: CreateCustomObject
    extra_custom_objects: Optional[List[Any]] = None
    # Type: CreateCustomResourceDefinition
    extra_crds: Optional[List[Any]] = None


class K8sApp(BaseApp):
    def __init__(self) -> None:
        super().__init__()

        # Args for the K8sApp, updated by the subclass
        self.args: K8sAppArgs = K8sAppArgs()

        # Dict of KubernetesResourceGroups
        # Type: Optional[Dict[str, K8sResourceGroup]]
        self.k8s_resource_groups: Optional[Dict[str, Any]] = None

    @property
    def container_name(self) -> str:
        from phidata.utils.k8s import get_default_container_name

        return self.args.container_name or get_default_container_name(self.name)

    @container_name.setter
    def container_name(self, container_name: str) -> None:
        if self.args is not None and container_name is not None:
            self.args.container_name = container_name

    @property
    def pod_name(self) -> str:
        from phidata.utils.k8s import get_default_pod_name

        return self.args.pod_name or get_default_pod_name(self.name)

    @pod_name.setter
    def pod_name(self, pod_name: str) -> None:
        if self.args is not None and pod_name is not None:
            self.args.pod_name = pod_name

    @property
    def secret_name(self) -> str:
        from phidata.utils.k8s import get_default_secret_name

        return self.args.secret_name or get_default_secret_name(self.name)

    @secret_name.setter
    def secret_name(self, secret_name: str) -> None:
        if self.args is not None and secret_name is not None:
            self.args.secret_name = secret_name

    @property
    def configmap_name(self) -> str:
        from phidata.utils.k8s import get_default_configmap_name

        return self.args.configmap_name or get_default_configmap_name(self.name)

    @configmap_name.setter
    def configmap_name(self, configmap_name: str) -> None:
        if self.args is not None and configmap_name is not None:
            self.args.configmap_name = configmap_name

    @property
    def deploy_name(self) -> str:
        from phidata.utils.k8s import get_default_deploy_name

        return self.args.deploy_name or get_default_deploy_name(self.name)

    @deploy_name.setter
    def deploy_name(self, deploy_name: str) -> None:
        if self.args is not None and deploy_name is not None:
            self.args.deploy_name = deploy_name

    @property
    def service_name(self) -> str:
        from phidata.utils.k8s import get_default_service_name

        return self.args.service_name or get_default_service_name(self.name)

    @service_name.setter
    def service_name(self, service_name: str) -> None:
        if self.args is not None and service_name is not None:
            self.args.service_name = service_name

    @property
    def service_port(self) -> int:
        return self.args.service_port

    @service_port.setter
    def service_port(self, service_port: int) -> None:
        if self.args is not None and service_port is not None:
            self.args.service_port = service_port

    @property
    def sa_name(self) -> str:
        from phidata.utils.k8s import get_default_sa_name

        return self.args.sa_name or get_default_sa_name(self.name)

    @sa_name.setter
    def sa_name(self, sa_name: str) -> None:
        if self.args is not None and sa_name is not None:
            self.args.sa_name = sa_name

    @property
    def cr_name(self) -> str:
        from phidata.utils.k8s import get_default_cr_name

        return self.args.cr_name or get_default_cr_name(self.name)

    @cr_name.setter
    def cr_name(self, cr_name: str) -> None:
        if self.args is not None and cr_name is not None:
            self.args.cr_name = cr_name

    @property
    def crb_name(self) -> str:
        from phidata.utils.k8s import get_default_crb_name

        return self.args.crb_name or get_default_crb_name(self.name)

    @crb_name.setter
    def crb_name(self, crb_name: str) -> None:
        if self.args is not None and crb_name is not None:
            self.args.crb_name = crb_name

    def get_cr_policy_rules(self) -> List[Any]:
        from phidata.k8s.create.rbac_authorization_k8s_io.v1.cluster_role import (
            PolicyRule,
        )

        return [
            PolicyRule(
                api_groups=[""],
                resources=["pods", "secrets", "configmaps"],
                verbs=["get", "list", "watch", "create", "update", "patch", "delete"],
            ),
            PolicyRule(
                api_groups=[""],
                resources=["pods/logs"],
                verbs=["get", "list", "watch"],
            ),
            PolicyRule(
                api_groups=[""],
                resources=["pods/exec"],
                verbs=["get", "create", "watch", "delete"],
            ),
        ]

    def get_container_env_k8s(
        self, container_paths: ContainerPathContext
    ) -> Dict[str, str]:
        from phidata.constants import (
            PYTHONPATH_ENV_VAR,
            PHIDATA_RUNTIME_ENV_VAR,
            SCRIPTS_DIR_ENV_VAR,
            STORAGE_DIR_ENV_VAR,
            META_DIR_ENV_VAR,
            PRODUCTS_DIR_ENV_VAR,
            NOTEBOOKS_DIR_ENV_VAR,
            WORKFLOWS_DIR_ENV_VAR,
            WORKSPACE_ROOT_ENV_VAR,
            WORKSPACE_CONFIG_DIR_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}
        container_env.update(
            {
                PHIDATA_RUNTIME_ENV_VAR: "kubernetes",
                SCRIPTS_DIR_ENV_VAR: container_paths.scripts_dir or "",
                STORAGE_DIR_ENV_VAR: container_paths.storage_dir or "",
                META_DIR_ENV_VAR: container_paths.meta_dir or "",
                PRODUCTS_DIR_ENV_VAR: container_paths.products_dir or "",
                NOTEBOOKS_DIR_ENV_VAR: container_paths.notebooks_dir or "",
                WORKFLOWS_DIR_ENV_VAR: container_paths.workflows_dir or "",
                WORKSPACE_ROOT_ENV_VAR: container_paths.workspace_root or "",
                WORKSPACE_CONFIG_DIR_ENV_VAR: container_paths.workspace_config_dir
                or "",
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": container_paths.requirements_file or "",
                "MOUNT_WORKSPACE": str(self.args.mount_workspace),
                "WORKSPACE_DIR_CONTAINER_PATH": str(
                    self.args.workspace_dir_container_path
                ),
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            }
        )

        if self.args.set_python_path:
            python_path = self.args.python_path
            if python_path is None:
                python_path = container_paths.workspace_root
                if self.args.add_python_paths is not None:
                    python_path = "{}:{}".format(
                        python_path, ":".join(self.args.add_python_paths)
                    )
            if python_path is not None:
                container_env[PYTHONPATH_ENV_VAR] = python_path

        # Set aws env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(
                {k: str(v) for k, v in env_data_from_file.items() if v is not None}
            )

        # Update the container env with user provided env
        # this overwrites any existing variables with the same key
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(
                {k: str(v) for k, v in self.args.env.items() if v is not None}
            )

        return container_env

    def get_container_labels_k8s(
        self, common_labels: Optional[Dict[str, str]]
    ) -> Dict[str, str]:
        labels: Dict[str, str] = common_labels or {}
        if self.args.container_labels is not None and isinstance(
            self.args.container_labels, dict
        ):
            labels.update(self.args.container_labels)
        return labels

    def get_deployment_labels_k8s(
        self, common_labels: Optional[Dict[str, str]]
    ) -> Dict[str, str]:
        labels: Dict[str, str] = common_labels or {}
        if self.args.container_labels is not None and isinstance(
            self.args.container_labels, dict
        ):
            labels.update(self.args.container_labels)
        return labels

    def get_service_labels_k8s(
        self, common_labels: Optional[Dict[str, str]]
    ) -> Dict[str, str]:
        labels: Dict[str, str] = common_labels or {}
        if self.args.container_labels is not None and isinstance(
            self.args.container_labels, dict
        ):
            labels.update(self.args.container_labels)
        return labels

    def get_container_args_k8s(self) -> Optional[List[str]]:
        if isinstance(self.args.command, str):
            return self.args.command.split(" ")
        return self.args.command

    def get_k8s_rg(self, k8s_build_context: Any) -> Optional[Any]:
        from phidata.k8s.create.common.port import CreatePort
        from phidata.k8s.create.core.v1.container import CreateContainer
        from phidata.k8s.create.core.v1.volume import (
            CreateVolume,
            HostPathVolumeSource,
            AwsElasticBlockStoreVolumeSource,
            VolumeType,
        )
        from phidata.k8s.create.group import (
            CreateK8sResourceGroup,
            CreateNamespace,
            CreateServiceAccount,
            CreateClusterRole,
            CreateClusterRoleBinding,
            CreateSecret,
            CreateConfigMap,
            CreateStorageClass,
            CreateService,
            CreateDeployment,
            CreateCustomObject,
            CreateCustomResourceDefinition,
            CreatePersistentVolume,
            CreatePVC,
        )
        from phidata.k8s.resource.group import K8sBuildContext
        from phidata.types.context import ContainerPathContext
        from phidata.utils.common import get_default_volume_name

        # -*- Build Container Paths
        container_paths: Optional[ContainerPathContext] = self.get_container_paths()
        if container_paths is None:
            raise Exception("Could not build Container Paths")
        # logger.debug(f"ContainerPaths: {container_paths.json(indent=2)}")

        app_name = self.name
        workspace_name = container_paths.workspace_name
        logger.debug(f"Building K8sResourceGroup: {app_name} for {workspace_name}")

        if k8s_build_context is None or not isinstance(
            k8s_build_context, K8sBuildContext
        ):
            logger.error("k8s_build_context must be a K8sBuildContext")
            return None

        # -*- Initialize K8s resources
        ns: Optional[CreateNamespace] = self.args.namespace
        sa: Optional[CreateServiceAccount] = self.args.service_account
        cr: Optional[CreateClusterRole] = self.args.cluster_role
        crb: Optional[CreateClusterRoleBinding] = self.args.cluster_role_binding
        secrets: List[CreateSecret] = self.args.extra_secrets or []
        config_maps: List[CreateConfigMap] = self.args.extra_configmaps or []
        services: List[CreateService] = self.args.extra_services or []
        deployments: List[CreateDeployment] = self.args.extra_deployments or []
        pvs: List[CreatePersistentVolume] = self.args.extra_pvs or []
        pvcs: List[CreatePVC] = self.args.extra_pvcs or []
        containers: List[CreateContainer] = self.args.extra_containers or []
        init_containers: List[CreateContainer] = self.args.extra_init_containers or []
        ports: List[CreatePort] = self.args.extra_ports or []
        volumes: List[CreateVolume] = self.args.extra_volumes or []
        storage_classes: List[CreateStorageClass] = (
            self.args.extra_storage_classes or []
        )
        custom_objects: List[CreateCustomObject] = self.args.extra_custom_objects or []
        crds: List[CreateCustomResourceDefinition] = self.args.extra_crds or []

        # Use the Namespace provided with the App or
        # use the default Namespace from the k8s_build_context
        ns_name: str = self.args.ns_name or k8s_build_context.namespace
        # Use the Service Account provided with the App or
        # use the default Service Account from the k8s_build_context
        sa_name: Optional[str] = (
            self.args.sa_name or k8s_build_context.service_account_name
        )
        common_labels: Optional[Dict[str, str]] = k8s_build_context.labels

        # -*- Build separate RBAC if use_rbac is True
        if self.args.use_rbac:
            # Create Namespace for this App
            if ns is None:
                ns = CreateNamespace(
                    ns=ns_name,
                    app_name=app_name,
                    labels=common_labels,
                )
            ns_name = ns.ns

            # Create Service Account for this App
            if sa is None:
                sa = CreateServiceAccount(
                    sa_name=sa_name or self.sa_name,
                    app_name=app_name,
                    namespace=ns_name,
                )
            sa_name = sa.sa_name

            # Create Cluster Role for this App
            if cr is None:
                cr = CreateClusterRole(
                    cr_name=self.cr_name,
                    rules=self.get_cr_policy_rules(),
                    app_name=app_name,
                    labels=common_labels,
                )

            # Create ClusterRoleBinding for this App
            if crb is None:
                crb = CreateClusterRoleBinding(
                    crb_name=self.crb_name,
                    cr_name=cr.cr_name,
                    service_account_name=sa.sa_name,
                    app_name=app_name,
                    namespace=ns_name,
                    labels=common_labels,
                )

        # -*- Build Container Environment
        container_env: Dict[str, str] = self.get_container_env_k8s(
            container_paths=container_paths
        )
        # Create a ConfigMap to set the Container Environment
        container_env_cm = CreateConfigMap(
            cm_name=self.configmap_name,
            app_name=app_name,
            namespace=ns_name,
            data=container_env,
            labels=common_labels,
        )
        config_maps.append(container_env_cm)

        # -*- Build Container Secrets
        container_secrets = self.get_secret_data()
        # Create a Secret to set the Container Secrets
        if container_secrets is not None:
            container_env_secret = CreateSecret(
                secret_name=self.secret_name,
                app_name=app_name,
                string_data=container_secrets,
                namespace=ns_name,
                labels=common_labels,
            )
            secrets.append(container_env_secret)

        # Add NodeSelectors to Pods for creating az sensitive volumes
        pod_node_selector: Optional[Dict[str, str]] = self.args.pod_node_selector
        # -*- Build Container Volumes
        if self.args.mount_workspace:
            workspace_volume_name = self.args.workspace_volume_name
            if workspace_volume_name is None:
                if workspace_name is not None:
                    workspace_volume_name = get_default_volume_name(
                        f"{app_name}-{workspace_name}-ws"
                    )
                else:
                    workspace_volume_name = get_default_volume_name(f"{app_name}-ws")
            # If workspace_volume_type is None or EmptyDir
            if (
                self.args.workspace_volume_type is None
                or self.args.workspace_volume_type == WorkspaceVolumeType.EmptyDir
            ):
                workspace_parent_container_path_str = container_paths.workspace_parent
                logger.debug(f"Creating EmptyDir")
                logger.debug(f"\tat: {workspace_parent_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_parent_container_path_str,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                volumes.append(workspace_volume)

                if self.args.create_git_sync_sidecar:
                    if self.args.git_sync_repo is not None:
                        git_sync_env = {
                            "GIT_SYNC_REPO": self.args.git_sync_repo,
                            "GIT_SYNC_ROOT": workspace_parent_container_path_str,
                            "GIT_SYNC_DEST": workspace_name,
                        }
                        if self.args.git_sync_branch is not None:
                            git_sync_env["GIT_SYNC_BRANCH"] = self.args.git_sync_branch
                        if self.args.git_sync_wait is not None:
                            git_sync_env["GIT_SYNC_WAIT"] = str(self.args.git_sync_wait)
                        git_sync_container = CreateContainer(
                            container_name="git-sync",
                            app_name=app_name,
                            image_name=self.args.git_sync_image_name,
                            image_tag=self.args.git_sync_image_tag,
                            env=git_sync_env,
                            envs_from_configmap=[cm.cm_name for cm in config_maps]
                            if len(config_maps) > 0
                            else None,
                            envs_from_secret=[secret.secret_name for secret in secrets]
                            if len(secrets) > 0
                            else None,
                            volumes=[workspace_volume],
                        )
                        containers.append(git_sync_container)

                        if self.args.create_git_sync_init_container:
                            git_sync_init_env: Dict[str, Any] = {
                                "GIT_SYNC_ONE_TIME": True
                            }
                            git_sync_init_env.update(git_sync_env)
                            _git_sync_init_container = CreateContainer(
                                container_name="git-sync-init",
                                app_name=git_sync_container.app_name,
                                image_name=git_sync_container.image_name,
                                image_tag=git_sync_container.image_tag,
                                env=git_sync_init_env,
                                envs_from_configmap=git_sync_container.envs_from_configmap,
                                envs_from_secret=git_sync_container.envs_from_secret,
                                volumes=git_sync_container.volumes,
                            )
                            init_containers.append(_git_sync_init_container)
                    else:
                        logger.error("GIT_SYNC_REPO invalid")
            # Use HostPath as WorkspaceVolumeType
            elif self.args.workspace_volume_type == WorkspaceVolumeType.HostPath:
                workspace_root_path_str = str(self.workspace_root_path)
                workspace_root_container_path_str = container_paths.workspace_root
                logger.debug(f"Mounting: {workspace_root_path_str}")
                logger.debug(f"\tto: {workspace_root_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_root_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=workspace_root_path_str,
                    ),
                )
                volumes.append(workspace_volume)

        if self.args.create_app_volume:
            app_volume_name = self.args.app_volume_name
            if app_volume_name is None:
                if workspace_name is not None:
                    volume_name = get_default_volume_name(
                        f"{app_name}-{workspace_name}"
                    )
                else:
                    volume_name = get_default_volume_name(app_name)
            # Use AwsEbs as AppVolumeType
            if self.args.app_volume_type == AppVolumeType.AwsEbs:
                if (
                    self.args.app_ebs_volume_id is not None
                    or self.args.app_ebs_volume is not None
                ):
                    # To use EbsVolume as the volume_type we:
                    # 1. Need the volume_id
                    # 2. Need to make sure pods are scheduled in the
                    #       same region/az as the volume

                    # For the volume_id we can either:
                    # 1. Use self.args.app_ebs_volume_id
                    # 2. OR get it from the self.args.app_ebs_volume
                    app_ebs_volume_id = self.args.app_ebs_volume_id

                    # For the aws_region/az we can either:
                    # 1. Use self.args.app_ebs_volume_region
                    # 2. OR get it from self.args.app_ebs_volume
                    # 3. OR get it from the BaseAppArgs
                    app_ebs_volume_region = self.args.app_ebs_volume_region
                    app_ebs_volume_az = self.args.app_ebs_volume_az

                    # Derive the aws_region from self.args.app_ebs_volume if needed
                    if (
                        app_ebs_volume_region is None
                        and self.args.app_ebs_volume is not None
                    ):
                        # Note: this will use the `$AWS_REGION` env var if set
                        _aws_region_from_ebs_volume = (
                            self.args.app_ebs_volume.get_aws_region()
                        )
                        if _aws_region_from_ebs_volume is not None:
                            app_ebs_volume_region = _aws_region_from_ebs_volume

                    # Get the aws_region from the BaseAppArgs
                    if app_ebs_volume_region is None:
                        app_ebs_volume_region = self.aws_region

                    # Derive the availability_zone from self.args.app_ebs_volume if needed
                    if (
                        app_ebs_volume_az is None
                        and self.args.app_ebs_volume is not None
                    ):
                        app_ebs_volume_az = self.args.app_ebs_volume.availability_zone

                    logger.debug(f"app_ebs_volume_region: {app_ebs_volume_region}")
                    logger.debug(f"app_ebs_volume_az: {app_ebs_volume_az}")

                    # Derive ebs_volume_id from self.args.app_ebs_volume if needed
                    if (
                        app_ebs_volume_id is None
                        and self.args.app_ebs_volume is not None
                    ):
                        app_ebs_volume_id = self.args.app_ebs_volume.get_volume_id(
                            aws_region=app_ebs_volume_region,
                            aws_profile=self.aws_profile,
                        )

                    logger.debug(f"app_ebs_volume_id: {app_ebs_volume_id}")
                    if app_ebs_volume_id is None:
                        logger.error("Could not find volume_id for AppEbsVolume")
                        return None

                    app_ebs_volume = CreateVolume(
                        volume_name=app_volume_name,
                        app_name=app_name,
                        mount_path=self.args.app_volume_container_path,
                        volume_type=VolumeType.AWS_EBS,
                        aws_ebs=AwsElasticBlockStoreVolumeSource(
                            volume_id=app_ebs_volume_id,
                        ),
                    )
                    volumes.append(app_ebs_volume)

                    # VERY IMPORTANT: pods should be scheduled in the same region/az as the volume
                    # To do this, we add NodeSelectors to Pods
                    if self.args.schedule_pods_in_ebs_topology:
                        if pod_node_selector is None:
                            pod_node_selector = {}

                        # Add NodeSelectors to Pods, so they are scheduled in the same
                        # region and zone as the ebs_volume
                        # https://kubernetes.io/docs/reference/labels-annotations-taints/#topologykubernetesiozone
                        if app_ebs_volume_region is not None:
                            pod_node_selector[
                                "topology.kubernetes.io/region"
                            ] = app_ebs_volume_region

                        if app_ebs_volume_az is not None:
                            pod_node_selector[
                                "topology.kubernetes.io/zone"
                            ] = app_ebs_volume_az
                else:
                    logger.error(f"{app_name}: app_ebs_volume not provided")
                    return None
            # Use EmptyDir as AppVolumeType
            elif self.args.app_volume_type == AppVolumeType.EmptyDir:
                empty_dir_volume = CreateVolume(
                    volume_name=app_volume_name,
                    app_name=app_name,
                    mount_path=self.args.app_volume_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                volumes.append(empty_dir_volume)
            # Use HostPath as AppVolumeType
            elif self.args.app_volume_type == AppVolumeType.HostPath:
                if self.args.app_volume_host_path is not None:
                    volume_host_path_str = str(self.args.app_volume_host_path)
                    host_path_volume = CreateVolume(
                        volume_name=app_volume_name,
                        app_name=app_name,
                        mount_path=self.args.app_volume_container_path,
                        volume_type=VolumeType.HOST_PATH,
                        host_path=HostPathVolumeSource(
                            path=volume_host_path_str,
                        ),
                    )
                    volumes.append(host_path_volume)
                else:
                    logger.error(f"{app_name}: app_volume_host_path not provided")
                    return None

        # -*- Build Container Ports
        if self.args.open_container_port:
            container_port = CreatePort(
                name=self.args.container_port_name,
                container_port=self.args.container_port,
                service_port=self.args.service_port,
                target_port=self.args.service_target_port
                or self.args.container_port_name,
            )
            ports.append(container_port)

        # -*- Build Container Labels
        container_labels: Dict[str, str] = self.get_container_labels_k8s(common_labels)

        # -*- Build Container Args: Equivalent to docker CMD
        container_args: Optional[List[str]] = self.get_container_args_k8s()
        if container_args:
            logger.debug("Command: {}".format(" ".join(container_args)))

        # -*- Create the Container
        container = CreateContainer(
            container_name=self.container_name,
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=container_args,
            # Equivalent to docker images ENTRYPOINT
            command=[self.args.entrypoint]
            if isinstance(self.args.entrypoint, str)
            else self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy
            or ImagePullPolicy.IF_NOT_PRESENT,
            envs_from_configmap=[cm.cm_name for cm in config_maps]
            if len(config_maps) > 0
            else None,
            envs_from_secret=[secret.secret_name for secret in secrets]
            if len(secrets) > 0
            else None,
            ports=ports if len(ports) > 0 else None,
            volumes=volumes if len(volumes) > 0 else None,
            labels=container_labels,
        )
        containers.insert(0, container)

        # Set default container for kubectl commands
        # https://kubernetes.io/docs/reference/labels-annotations-taints/#kubectl-kubernetes-io-default-container
        pod_annotations = {
            "kubectl.kubernetes.io/default-container": container.container_name
        }

        # -*- Add pod annotations
        if self.args.pod_annotations is not None and isinstance(
            self.args.pod_annotations, dict
        ):
            pod_annotations.update(self.args.pod_annotations)

        # -*- Build Deployment Labels
        deploy_labels: Dict[str, str] = self.get_deployment_labels_k8s(common_labels)

        # If using EbsVolume, restart the deployment on update
        recreate_deployment_on_update = (
            True
            if (
                self.args.create_app_volume
                and self.args.app_volume_type == AppVolumeType.AwsEbs
            )
            else False
        )

        # -*- Create the Deployment
        deployment = CreateDeployment(
            deploy_name=self.deploy_name,
            pod_name=self.pod_name,
            app_name=app_name,
            namespace=ns_name,
            service_account_name=sa_name,
            replicas=self.args.replicas,
            containers=containers,
            init_containers=init_containers if len(init_containers) > 0 else None,
            pod_node_selector=pod_node_selector,
            restart_policy=self.args.deploy_restart_policy or RestartPolicy.ALWAYS,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=volumes if len(volumes) > 0 else None,
            labels=deploy_labels,
            pod_annotations=pod_annotations,
            topology_spread_key=self.args.topology_spread_key,
            topology_spread_max_skew=self.args.topology_spread_max_skew,
            topology_spread_when_unsatisfiable=self.args.topology_spread_when_unsatisfiable,
            recreate_on_update=recreate_deployment_on_update,
        )
        deployments.append(deployment)

        # -*- Create the Service
        if self.args.create_service:
            service_labels: Dict[str, str] = self.get_service_labels_k8s(common_labels)
            _service = CreateService(
                service_name=self.service_name,
                app_name=app_name,
                namespace=ns_name,
                service_account_name=sa_name,
                service_type=self.args.service_type,
                deployment=deployment,
                ports=ports if len(ports) > 0 else None,
                labels=service_labels,
            )
            services.append(_service)

        # -*- Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            ns=ns,
            sa=sa,
            cr=cr,
            crb=crb,
            secrets=secrets if len(secrets) > 0 else None,
            config_maps=config_maps if len(config_maps) > 0 else None,
            storage_classes=storage_classes if len(storage_classes) > 0 else None,
            services=services if len(services) > 0 else None,
            deployments=deployments if len(deployments) > 0 else None,
            custom_objects=custom_objects if len(custom_objects) > 0 else None,
            crds=crds if len(crds) > 0 else None,
            pvs=pvs if len(pvs) > 0 else None,
            pvcs=pvcs if len(pvcs) > 0 else None,
        )

        return k8s_resource_group.create()

    def build_k8s_resource_groups(self, docker_build_context: Any) -> None:
        k8s_rg = self.get_k8s_rg(docker_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg

    def get_k8s_resource_groups(
        self, k8s_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Dict[str, Any]]:
        if self.k8s_resource_groups is None:
            self.build_k8s_resource_groups(k8s_build_context)
        # # Comment out in production
        # if self.k8s_resource_groups:
        #     logger.debug("K8sResourceGroups:")
        #     for rg_name, rg in self.k8s_resource_groups.items():
        #         logger.debug(
        #             "{}:{}\n{}".format(rg_name, type(rg), rg)
        #         )
        return self.k8s_resource_groups



import time
from typing import Any
from phidata.utils.log import logger

try:
    from langchain.tools import BaseTool
except ImportError:
    logger.warning("Langchain not available.")
    raise

from phidata.llm.duckdb.query import run_duckdb_query


class DuckDBQueryTool(BaseTool):
    name: str = "execute"
    description: str = """Useful for when you need to run SQL queries against a DuckDB database.
    Input to this tool is a detailed and correct SQL query, output is a result from the database.
    If the query is not correct, an error message will be returned.
    If an error is returned, rewrite the query, check the query, and try again.
    """
    duckdb_connection: Any = None

    def _run(self, query: str) -> str:
        """Use the tool to run a query against the database."""
        query_result = run_duckdb_query(self.duckdb_connection, query)
        time.sleep(1)
        return query_result

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("Not supported")



from typing import Optional, Any, List, Dict
from collections import OrderedDict

from phidata.app.base_app import BaseApp, BaseAppArgs
from phidata.types.context import ContainerPathContext
from phidata.utils.log import logger


class AwsAppArgs(BaseAppArgs):
    # -*- AWS Configuration
    # List of subnets: str or Subnet
    aws_subnets: Optional[List[Any]] = None
    # List of security groups: str or SecurityGroup
    aws_security_groups: Optional[List[Any]] = None

    # -*- ECS Configuration
    ecs_cluster: Optional[Any] = None
    ecs_launch_type: str = "FARGATE"
    ecs_task_cpu: str = "1024"
    ecs_task_memory: str = "2048"
    ecs_service_count: int = 1
    assign_public_ip: Optional[bool] = None
    ecs_enable_exec: bool = True

    # -*- LoadBalancer Configuration
    load_balancer: Optional[Any] = None
    listener: Optional[Any] = None
    # Create a load balancer if load_balancer is None
    create_load_balancer: bool = False
    # HTTP or HTTPS
    load_balancer_protocol: str = "HTTP"
    load_balancer_security_groups: Optional[List[Any]] = None
    # Default 80 for HTTP and 443 for HTTPS
    load_balancer_port: Optional[int] = None
    load_balancer_certificate: Optional[Any] = None
    load_balancer_certificate_arn: Optional[str] = None

    # -*- TargetGroup Configuration
    target_group: Optional[Any] = None
    # HTTP or HTTPS
    target_group_protocol: str = "HTTP"
    # Default 80 for HTTP and 443 for HTTPS
    target_group_port: Optional[int] = None
    target_group_type: str = "ip"
    health_check_protocol: Optional[str] = None
    health_check_port: Optional[str] = None
    health_check_enabled: Optional[bool] = None
    health_check_path: Optional[str] = None
    health_check_interval_seconds: Optional[int] = None
    health_check_timeout_seconds: Optional[int] = None
    healthy_threshold_count: Optional[int] = None
    unhealthy_threshold_count: Optional[int] = None


class AwsApp(BaseApp):
    def __init__(self) -> None:
        super().__init__()

        # Args for the AwsAppArgs, updated by the subclass
        self.args: AwsAppArgs = AwsAppArgs()

        # Dict of AwsResourceGroups
        # Type: Optional[Dict[str, AwsResourceGroup]]
        self.aws_resource_groups: Optional[Dict[str, Any]] = None

    @property
    def aws_subnets(self) -> Optional[List[Any]]:
        return self.args.aws_subnets

    @aws_subnets.setter
    def aws_subnets(self, aws_subnets: List[Any]) -> None:
        if self.args is not None and aws_subnets is not None:
            self.args.aws_subnets = aws_subnets

    @property
    def aws_security_groups(self) -> Optional[List[Any]]:
        return self.args.aws_security_groups

    @aws_security_groups.setter
    def aws_security_groups(self, aws_security_groups: List[Any]) -> None:
        if self.args is not None and aws_security_groups is not None:
            self.args.aws_security_groups = aws_security_groups

    @property
    def ecs_cluster(self) -> Optional[Any]:
        return self.args.ecs_cluster

    @ecs_cluster.setter
    def ecs_cluster(self, ecs_cluster: Any) -> None:
        if self.args is not None and ecs_cluster is not None:
            self.args.ecs_cluster = ecs_cluster

    @property
    def ecs_launch_type(self) -> Optional[str]:
        return self.args.ecs_launch_type

    @ecs_launch_type.setter
    def ecs_launch_type(self, ecs_launch_type: str) -> None:
        if self.args is not None and ecs_launch_type is not None:
            self.args.ecs_launch_type = ecs_launch_type

    @property
    def ecs_task_cpu(self) -> Optional[str]:
        return self.args.ecs_task_cpu

    @ecs_task_cpu.setter
    def ecs_task_cpu(self, ecs_task_cpu: str) -> None:
        if self.args is not None and ecs_task_cpu is not None:
            self.args.ecs_task_cpu = ecs_task_cpu

    @property
    def ecs_task_memory(self) -> Optional[str]:
        return self.args.ecs_task_memory

    @ecs_task_memory.setter
    def ecs_task_memory(self, ecs_task_memory: str) -> None:
        if self.args is not None and ecs_task_memory is not None:
            self.args.ecs_task_memory = ecs_task_memory

    @property
    def ecs_service_count(self) -> Optional[int]:
        return self.args.ecs_service_count

    @ecs_service_count.setter
    def ecs_service_count(self, ecs_service_count: int) -> None:
        if self.args is not None and ecs_service_count is not None:
            self.args.ecs_service_count = ecs_service_count

    @property
    def assign_public_ip(self) -> Optional[bool]:
        return self.args.assign_public_ip

    @assign_public_ip.setter
    def assign_public_ip(self, assign_public_ip: bool) -> None:
        if self.args is not None and assign_public_ip is not None:
            self.args.assign_public_ip = assign_public_ip

    @property
    def ecs_enable_exec(self) -> Optional[bool]:
        return self.args.ecs_enable_exec

    @ecs_enable_exec.setter
    def ecs_enable_exec(self, ecs_enable_exec: bool) -> None:
        if self.args is not None and ecs_enable_exec is not None:
            self.args.ecs_enable_exec = ecs_enable_exec

    @property
    def create_load_balancer(self) -> Optional[bool]:
        return self.args.create_load_balancer

    @create_load_balancer.setter
    def create_load_balancer(self, create_load_balancer: bool) -> None:
        if self.args is not None and create_load_balancer is not None:
            self.args.create_load_balancer = create_load_balancer

    @property
    def load_balancer(self) -> Optional[Any]:
        return self.args.load_balancer

    @load_balancer.setter
    def load_balancer(self, load_balancer: Any) -> None:
        if self.args is not None and load_balancer is not None:
            self.args.load_balancer = load_balancer

    @property
    def load_balancer_protocol(self) -> Optional[str]:
        return self.args.load_balancer_protocol

    @load_balancer_protocol.setter
    def load_balancer_protocol(self, load_balancer_protocol: str) -> None:
        if self.args is not None and load_balancer_protocol is not None:
            self.args.load_balancer_protocol = load_balancer_protocol

    @property
    def load_balancer_port(self) -> Optional[int]:
        return self.args.load_balancer_port

    @load_balancer_port.setter
    def load_balancer_port(self, load_balancer_port: int) -> None:
        if self.args is not None and load_balancer_port is not None:
            self.args.load_balancer_port = load_balancer_port

    @property
    def load_balancer_certificate_arn(self) -> Optional[str]:
        return self.args.load_balancer_certificate_arn

    @load_balancer_certificate_arn.setter
    def load_balancer_certificate_arn(self, load_balancer_certificate_arn: str) -> None:
        if self.args is not None and load_balancer_certificate_arn is not None:
            self.args.load_balancer_certificate_arn = load_balancer_certificate_arn

    def get_container_env_ecs(
        self, container_paths: ContainerPathContext
    ) -> Dict[str, str]:
        from phidata.constants import (
            PYTHONPATH_ENV_VAR,
            PHIDATA_RUNTIME_ENV_VAR,
            SCRIPTS_DIR_ENV_VAR,
            STORAGE_DIR_ENV_VAR,
            META_DIR_ENV_VAR,
            PRODUCTS_DIR_ENV_VAR,
            NOTEBOOKS_DIR_ENV_VAR,
            WORKFLOWS_DIR_ENV_VAR,
            WORKSPACE_ROOT_ENV_VAR,
            WORKSPACE_CONFIG_DIR_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}
        container_env.update(
            {
                PHIDATA_RUNTIME_ENV_VAR: "ecs",
                SCRIPTS_DIR_ENV_VAR: container_paths.scripts_dir or "",
                STORAGE_DIR_ENV_VAR: container_paths.storage_dir or "",
                META_DIR_ENV_VAR: container_paths.meta_dir or "",
                PRODUCTS_DIR_ENV_VAR: container_paths.products_dir or "",
                NOTEBOOKS_DIR_ENV_VAR: container_paths.notebooks_dir or "",
                WORKFLOWS_DIR_ENV_VAR: container_paths.workflows_dir or "",
                WORKSPACE_ROOT_ENV_VAR: container_paths.workspace_root or "",
                WORKSPACE_CONFIG_DIR_ENV_VAR: container_paths.workspace_config_dir
                or "",
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": container_paths.requirements_file or "",
                "MOUNT_WORKSPACE": str(self.args.mount_workspace),
                "WORKSPACE_DIR_CONTAINER_PATH": str(
                    self.args.workspace_dir_container_path
                ),
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            }
        )

        if self.args.set_python_path:
            python_path = self.args.python_path
            if python_path is None:
                python_path = container_paths.workspace_root
                if self.args.add_python_paths is not None:
                    python_path = "{}:{}".format(
                        python_path, ":".join(self.args.add_python_paths)
                    )
            if python_path is not None:
                container_env[PYTHONPATH_ENV_VAR] = python_path

        # Set aws env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(
                {k: str(v) for k, v in env_data_from_file.items() if v is not None}
            )

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data()
        if secret_data_from_file is not None:
            container_env.update(
                {k: str(v) for k, v in secret_data_from_file.items() if v is not None}
            )

        # Update the container env with user provided env
        # this overwrites any existing variables with the same key
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(
                {k: v for k, v in self.args.env.items() if v is not None}
            )

        # logger.debug("Container Environment: {}".format(container_env))
        return container_env

    def get_container_command_aws(self) -> Optional[List[str]]:
        if isinstance(self.args.command, str):
            return self.args.command.split(" ")
        return self.args.command

    def get_aws_rg(
        self, aws_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Any]:
        from phidata.aws.resource.group import (
            AwsResourceGroup,
            EcsCluster,
            EcsContainer,
            EcsTaskDefinition,
            EcsService,
            LoadBalancer,
            TargetGroup,
            Listener,
            AcmCertificate,
            SecurityGroup,
        )

        # -*- Build Container Paths
        container_paths: Optional[ContainerPathContext] = self.get_container_paths()
        if container_paths is None:
            raise Exception("Could not build Container Paths")
        logger.debug(f"ContainerPaths: {container_paths.json(indent=2)}")

        app_name = self.name
        workspace_name = container_paths.workspace_name
        logger.debug(f"Building AwsResourceGroup: {app_name} for {workspace_name}")

        # -*- Build Container Environment
        container_env: Dict[str, str] = self.get_container_env_ecs(
            container_paths=container_paths
        )

        # -*- Create Security Groups
        security_groups: List[SecurityGroup] = []
        if self.args.aws_security_groups is not None:
            for sg in self.args.aws_security_groups:
                if isinstance(sg, SecurityGroup):
                    security_groups.append(sg)
        if self.args.load_balancer_security_groups is not None:
            for lb_sg in self.args.load_balancer_security_groups:
                if isinstance(lb_sg, SecurityGroup):
                    security_groups.append(lb_sg)

        # -*- Create ECS cluster
        ecs_cluster = self.args.ecs_cluster
        if ecs_cluster is None:
            ecs_cluster = EcsCluster(
                name=f"{app_name}-cluster",
                ecs_cluster_name=app_name,
                capacity_providers=[self.args.ecs_launch_type],
                save_output=self.args.save_output,
                resource_dir=self.args.resource_dir or app_name,
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )

        # -*- Create Load Balancer
        load_balancer = self.args.load_balancer
        if load_balancer is None and self.args.create_load_balancer:
            if self.args.load_balancer_protocol not in ["HTTP", "HTTPS"]:
                raise Exception(
                    "Load Balancer Protocol must be one of: HTTP, HTTPS. "
                    f"Got: {self.args.load_balancer_protocol}"
                )
            load_balancer_sgs = (
                self.args.load_balancer_security_groups or self.args.aws_security_groups
            )
            load_balancer = LoadBalancer(
                name=f"{app_name}-lb",
                subnets=self.args.aws_subnets,
                security_groups=load_balancer_sgs,
                protocol=self.args.load_balancer_protocol,
                save_output=self.args.save_output,
                resource_dir=self.args.resource_dir or app_name,
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )

        # -*- Create Target Group
        target_group = self.args.target_group
        if target_group is None and self.args.create_load_balancer:
            if self.args.target_group_protocol not in ["HTTP", "HTTPS"]:
                raise Exception(
                    "Target Group Protocol must be one of: HTTP, HTTPS. "
                    f"Got: {self.args.target_group_protocol}"
                )
            target_group = TargetGroup(
                name=f"{app_name}-tg",
                port=self.container_port,
                protocol=self.args.target_group_protocol,
                subnets=self.args.aws_subnets,
                target_type=self.args.target_group_type,
                health_check_protocol=self.args.health_check_protocol,
                health_check_port=self.args.health_check_port,
                health_check_enabled=self.args.health_check_enabled,
                health_check_path=self.args.health_check_path,
                health_check_interval_seconds=self.args.health_check_interval_seconds,
                health_check_timeout_seconds=self.args.health_check_timeout_seconds,
                healthy_threshold_count=self.args.healthy_threshold_count,
                unhealthy_threshold_count=self.args.unhealthy_threshold_count,
                save_output=self.args.save_output,
                resource_dir=self.args.resource_dir or app_name,
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )

        # -*- Create Listener
        listener = self.args.listener
        if listener is None and self.args.create_load_balancer:
            listener = Listener(
                name=f"{app_name}-listener",
                load_balancer=load_balancer,
                target_group=target_group,
                save_output=self.args.save_output,
                resource_dir=self.args.resource_dir or app_name,
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )
            if self.args.load_balancer_certificate_arn is not None:
                listener.certificates = [
                    {"CertificateArn": self.args.load_balancer_certificate_arn}
                ]
            if self.args.load_balancer_certificate is not None:
                listener.acm_certificates = [self.args.load_balancer_certificate]

        # -*- Build Container Command
        container_cmd: Optional[List[str]] = self.get_container_command_aws()
        if container_cmd:
            logger.debug("Command: {}".format(" ".join(container_cmd)))

        # -*- Create ECS Container
        ecs_container = EcsContainer(
            name=app_name,
            image=self.get_image_str(),
            port_mappings=[{"containerPort": self.container_port}],
            command=container_cmd,
            environment=[{"name": k, "value": v} for k, v in container_env.items()],
            log_configuration={
                "logDriver": "awslogs",
                "options": {
                    "awslogs-group": app_name,
                    "awslogs-region": self.aws_region,
                    "awslogs-create-group": "true",
                    "awslogs-stream-prefix": app_name,
                },
            },
            env_from_secrets=self.args.aws_secrets,
            save_output=self.args.save_output,
            resource_dir=self.args.resource_dir or app_name,
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )

        # -*- Create ECS Task Definition
        ecs_task_definition = EcsTaskDefinition(
            name=f"{app_name}-td",
            family=app_name,
            network_mode="awsvpc",
            cpu=self.args.ecs_task_cpu,
            memory=self.args.ecs_task_memory,
            containers=[ecs_container],
            requires_compatibilities=[self.args.ecs_launch_type],
            add_ecs_exec_policy=True,
            add_ecs_secret_policy=True,
            save_output=self.args.save_output,
            resource_dir=self.args.resource_dir or app_name,
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )

        # -*- Create ECS Service
        ecs_service = EcsService(
            name=f"{app_name}-service",
            desired_count=self.args.ecs_service_count,
            launch_type=self.args.ecs_launch_type,
            cluster=ecs_cluster,
            task_definition=ecs_task_definition,
            target_group=target_group,
            target_container_name=ecs_container.name,
            target_container_port=self.container_port,
            subnets=self.args.aws_subnets,
            security_groups=self.args.aws_security_groups,
            assign_public_ip=self.args.assign_public_ip,
            # Force delete the service.
            force_delete=True,
            # Force a new deployment of the service on update.
            force_new_deployment=True,
            save_output=self.args.save_output,
            resource_dir=self.args.resource_dir or app_name,
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )

        # -*- Create AwsResourceGroup
        return AwsResourceGroup(
            name=app_name,
            enabled=self.enabled,
            ecs_clusters=[ecs_cluster],
            ecs_task_definitions=[ecs_task_definition],
            ecs_services=[ecs_service],
            load_balancers=[load_balancer],
            target_groups=[target_group],
            listeners=[listener],
            security_groups=security_groups if len(security_groups) > 0 else None,
        )

    def build_aws_resource_groups(
        self, aws_build_context: Any, defer_api_calls: bool = False
    ) -> None:
        aws_rg = self.get_aws_rg(aws_build_context)
        if aws_rg is not None:
            if self.aws_resource_groups is None:
                self.aws_resource_groups = OrderedDict()
            self.aws_resource_groups[aws_rg.name] = aws_rg

    def get_aws_resource_groups(
        self, aws_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Dict[str, Any]]:
        if self.aws_resource_groups is None:
            self.build_aws_resource_groups(aws_build_context)
        # # Comment out in production
        # if self.aws_resource_groups:
        #     logger.debug("AwsResourceGroups:")
        #     for rg_name, rg in self.aws_resource_groups.items():
        #         logger.debug(
        #             "{}:{}\n{}".format(rg_name, type(rg), rg)
        #         )
        return self.aws_resource_groups



import textwrap
from phidata.utils.log import logger

try:
    import duckdb
except ImportError:
    logger.warning("DuckDB not available.")
    raise

try:
    from langchain import LLMMathChain, SerpAPIWrapper
    from langchain.agents import Tool, initialize_agent
    from langchain.llms import OpenAIChat
except ImportError:
    logger.warning("Langchain not available.")
    raise

from phidata.llm.duckdb.chain import get_duckdb_query_chain
from phidata.llm.duckdb.query import run_duckdb_query, describe_table_or_view


def create_duckdb_agent(
    duckdb_connection: duckdb.DuckDBPyConnection, model_name: str = "gpt-3.5-turbo"
):
    """
    Agents use an LLM to determine which actions to take and in what order.
    An action can either be using a tool and observing its output, or returning to the user.

    This function creates an agent that can answer questions about a duckdb database.
    """

    # Step 1: Load the LLM to control the agent.
    llm = OpenAIChat(model_name=model_name, temperature=0)  # type: ignore

    # Step 2: Load tools that the agent can use to answer questions.
    llm_math_chain = LLMMathChain(llm=llm, verbose=True)
    db_op_chain = get_duckdb_query_chain(llm=llm, duckdb_connection=duckdb_connection)
    tools = [
        Tool(
            name="Calculator",
            func=llm_math_chain.run,
            description="Useful for when you need to answer questions about math",
        ),
        Tool(
            name="Show Tables",
            func=lambda _: run_duckdb_query(duckdb_connection, "show tables;"),
            description="Useful to show the available tables and views. Empty input required.",
        ),
        Tool(
            name="Describe Table",
            func=lambda table: describe_table_or_view(duckdb_connection, table),
            description="Useful to show the column names and types of a table or view. Use the table name as the input.",  # noqa: E501
        ),
        Tool(
            name="Data Op",
            func=lambda input: db_op_chain(
                {
                    "table_names": lambda _: run_duckdb_query(
                        duckdb_connection, "show tables;"
                    ),
                    "input": input,
                }
            ),
            description=textwrap.dedent(
                """useful for when you need to operate on data and answer questions by querying for data.
            Input should be in the form of a natural language question containing full context
            including what tables and columns are relevant to the question. Use only after data is present and loaded.
            """,  # noqa: E501
            ),
        ),
    ]

    # Step 2: Initialize the agent
    agent = initialize_agent(
        tools,
        llm,
        agent="zero-shot-react-description",  # type: ignore
        agent_kwargs={
            "input_variables": ["input", "agent_scratchpad", "table_names"],
            "prefix": prompt_prefix,
            "suffix": prompt_suffix,
        },
        # return_intermediate_steps=True,
        verbose=True,
    )
    return agent


prompt_prefix = """Answer the following question as best you can by querying for data to back up
your answer. Even if you know the answer, you MUST show you can get the answer from the database.

Refuse to delete any data, or drop tables. When answering, you MUST query the database for any data.
Check the available tables exist first. Prefer to take single independent actions. Prefer to create views
of data as one action, then select data from the view.

Always share the SQL queries you use to get the answer.

It is important that you use the exact phrase "Final Answer: " in your final answer.
List all SQL queries returned by Data Op in your final answer.

DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.

If the question does not seem related to the database, just return "I don't know" as the answer.

You have access to the following data tables:
{table_names}

Only use the below tools. You have access to the following tools:
"""

prompt_suffix = """
It is important that you use the exact phrase "Final Answer: <Summary>" in your final answer.

Begin! Remember to share the SQL queries you use to get the answer.

Question: {input}
Thought: I should look at the tables in the database to see what I can query.
{agent_scratchpad}"""
